"""Nueron node."""

from typing import Generic, Iterable, Iterator, List, overload

import numpy as np
import numpy.typing as npt

from .node import Node
from .swc import DictSWC, SWCLike, SWCTypeVar

__all__ = ["Path"]


class Path(SWCLike, Generic[SWCTypeVar]):
    """Neural path.

    A path is a linear set of points without bifurcations.
    """

    attach: SWCTypeVar
    idx: npt.NDArray[np.int32]

    class Node(Node["Path"]):
        """Node of neuron tree."""

    def __init__(self, attach: SWCTypeVar, idx: npt.ArrayLike) -> None:
        super().__init__()
        self.attach = attach
        self.idx = np.array(idx, dtype=np.int32)

    def __iter__(self) -> Iterator[Node]:
        return (self[i] for i in range(len(self)))

    def __len__(self) -> int:
        return self.id().shape[0]

    def __repr__(self) -> str:
        return f"Neuron path with {len(self)} nodes."

    # fmt:off
    @overload
    def __getitem__(self, key: int) -> Node: ...
    @overload
    def __getitem__(self, key: slice) -> List[Node]: ...
    @overload
    def __getitem__(self, key: str) -> npt.NDArray: ...
    # fmt:on

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [self.get_node(i) for i in range(*key.indices(len(self)))]

        if isinstance(key, int):
            length = len(self)
            if key < -length or key >= length:
                raise IndexError(f"The index ({key}) is out of range.")

            if key < 0:  # Handle negative indices
                key += length

            return self.get_node(key)

        if isinstance(key, str):
            return self.get_ndata(key)

        raise TypeError("Invalid argument type.")

    def keys(self) -> Iterable[str]:
        return self.attach.keys()

    def get_ndata(self, key: str) -> npt.NDArray:
        return self.attach.get_ndata(key)[self.idx]

    def get_node(self, idx: int) -> Node:
        return self.Node(self, idx)

    def detach(self) -> "Path":
        """Detach from current attached object."""
        attact = DictSWC(**{k: self[k] for k in self.keys()})
        return Path(attact, self.idx.copy())

    def id(self) -> npt.NDArray[np.int32]:  # pylint: disable=invalid-name
        """Get the ids of shape (n_sample,).

        Returns a consecutively incremented id.

        See Also
        --------
        self.origin_id
        """
        return np.arange(len(self.origin_id()), dtype=np.int32)

    def pid(self) -> npt.NDArray[np.int32]:
        """Get the ids of shape (n_sample,).

        Returns a consecutively incremented pid.

        See Also
        --------
        self.origin_pid
        """
        return np.arange(-1, len(self.origin_id()) - 1, dtype=np.int32)

    def origin_id(self) -> npt.NDArray[np.int32]:
        """Get the original id."""
        return self.get_ndata("id")

    def origin_pid(self) -> npt.NDArray[np.int32]:
        """Get the original pid."""
        return self.get_ndata("pid")

    def length(self) -> float:
        """Sum of length of stems."""
        xyz = self.xyz()
        return np.sum(np.linalg.norm(xyz[1:] - xyz[:-1], axis=1)).item()

    def straight_line_distance(self) -> float:
        """Straight-line distance of path.

        The end-to-end straight-line distance between start point and
        end point.
        """
        return np.linalg.norm(self[-1].xyz() - self[0].xyz()).item()

    def tortuosity(self) -> float:
        """Tortuosity of path.

        The straight-line distance between two consecutive branch
        points divided by the length of the neuronal path between
        those points.
        """
        if (length := self.length()) == 0:
            return 1
        return self.straight_line_distance() / length



"""Dgl grapth transforms."""

from typing import List, cast

import dgl

from ...core import Tree
from .. import Transform

__all__ = ["ToDGLGraph"]


class ToDGLGraph(Transform[Tree, dgl.DGLGraph]):
    """Transofrm tree to dgl graph.

    Notes
    -----
    You SHOULD initially decide to constrct your own transformation,
    even if we are attempting to provide additional capabilities for
    you because there are so many custom initial options in graph and
    this class is more of a toy and template.
    """

    keys: List[str] | None
    to_bidirected: bool

    def __init__(
        self,
        to_bidirected: bool = False,
        keys: List[str] | None = None,
    ) -> None:
        """Transofrm tree to dgl graph.

        Parameters
        ----------
        to_bidirected : bool, default to `False`
            If True, return bidirected graph.
        keys : List[str], optional
            Copy these keys as ndata of graph.
        """

        self.to_bidirected = to_bidirected
        self.keys = keys

    def __call__(self, x: Tree) -> dgl.DGLGraph:
        g = dgl.graph((x.id()[1:], x.pid()[1:]), num_nodes=x.number_of_nodes())
        if self.to_bidirected:
            g = cast(dgl.DGLGraph, dgl.to_bidirected(g))

        if self.keys:
            for k in self.keys:
                g.ndata[k] = x[k]

        return g

    def __repr__(self) -> str:
        return (
            "ToDGLGraph"
            + ("-ToBidirected" if self.to_bidirected else "")
            + (("-" + "-".join(self.keys)) if self.keys else "")
        )



"""Transformation in branch."""

from typing import cast

import numpy as np
import numpy.typing as npt

from ..core import Branch
from ..utils import (
    angle,
    rotate3d_x,
    rotate3d_y,
    rotate3d_z,
    scale3d,
    to_homogeneous,
    translate3d,
)
from .base import Transform

__all__ = ["BranchLinearResampler", "BranchStandardizer"]


class _BranchResampler(Transform[Branch, Branch]):
    r"""Resample branch."""

    def __call__(self, x: Branch) -> Branch:
        xyzr = x.xyzr()
        new_xyzr = self.resample(xyzr)
        return Branch.from_xyzr(new_xyzr)

    def resample(self, xyzr: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        raise NotImplementedError()


class BranchLinearResampler(_BranchResampler):
    r"""Resampling by linear interpolation, DO NOT keep original node."""

    def __init__(self, n_nodes: int) -> None:
        """Resample branch to special num of nodes.

        Parameters
        ----------
        n_nodes : int
            Number of nodes after resample.
        """
        super().__init__()
        self.n_nodes = n_nodes

    def __repr__(self) -> str:
        return f"BranchLinearResampler-{self.n_nodes}"

    def resample(self, xyzr: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        """Resampling by linear interpolation, DO NOT keep original node.

        Parameters
        ----------
        xyzr : np.ndarray[np.float32]
            The array of shape (N, 4).

        Returns
        -------
        coordinates : ~numpy.NDArray[float64]
            An array of shape (n_nodes, 4).
        """

        xp = np.cumsum(np.linalg.norm(xyzr[1:, :3] - xyzr[:-1, :3], axis=1))
        xp = np.insert(xp, 0, 0)
        xvals = np.linspace(0, xp[-1], self.n_nodes)

        x = np.interp(xvals, xp, xyzr[:, 0])
        y = np.interp(xvals, xp, xyzr[:, 1])
        z = np.interp(xvals, xp, xyzr[:, 2])
        r = np.interp(xvals, xp, xyzr[:, 3])
        return cast(npt.NDArray[np.float32], np.stack([x, y, z, r], axis=1))


class BranchStandardizer(Transform[Branch, Branch]):
    r"""Standarize branch.

    Standardized branch starts at (0, 0, 0), ends at (1, 0, 0), up at y,
    and scale max radius to 1.
    """

    def __call__(self, x: Branch) -> Branch:
        xyzr = x.xyzr()
        xyz, r = xyzr[:, 0:3], xyzr[:, 3:4]
        T = self.get_matrix(xyz)

        xyz4 = to_homogeneous(xyz, 1).transpose()  # (4, N)
        new_xyz = np.dot(T, xyz4)[:3].transpose()
        new_xyzr = np.concatenate([new_xyz, r / r.max()], axis=1)
        return Branch.from_xyzr(new_xyzr)

    @staticmethod
    def get_matrix(xyz: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        r"""Get standarize transformation matrix.

        Standardized branch starts at (0, 0, 0), ends at (1, 0, 0), up at y.

        Parameters
        ----------
        xyz : np.ndarray[np.float32]
            The `x`, `y`, `z` matrix of shape (N, 3) of branch.

        Returns
        -------
        T : np.ndarray[np.float32]
            An homogeneous transfomation matrix of shape (4, 4).
        """

        assert (
            xyz.ndim == 2 and xyz.shape[1] == 3
        ), f"xyz should be of shape (N, 3), got {xyz.shape}"

        xyz = xyz[:, :3]
        T = np.identity(4)
        v = np.concatenate([xyz[-1] - xyz[0], np.zeros((1))])[:, None]

        # translate to the origin
        T = translate3d(-xyz[0, 0], -xyz[0, 1], -xyz[0, 2]).dot(T)

        # scale to unit vector
        s = (1 / np.linalg.norm(v[:3, 0])).item()
        T = scale3d(s, s, s).dot(T)

        # rotate v to the xz-plane, v should be (x, 0, z) now
        vy = np.dot(T, v)[:, 0]
        # when looking at the xz-plane along the positive y-axis, the
        # coordinates should be (z, x)
        T = rotate3d_y(angle([vy[2], vy[0]], [0, 1])).dot(T)

        # rotate v to the x-axis, v should be (1, 0, 0) now
        vx = np.dot(T, v)[:, 0]
        T = rotate3d_z(angle([vx[0], vx[1]], [1, 0])).dot(T)

        # rotate the farthest point to the xy-plane
        if xyz.shape[0] > 2:
            xyz4 = to_homogeneous(xyz, 1).transpose()  # (4, N)
            new_xyz4 = np.dot(T, xyz4)  # (4, N)
            max_index = np.argmax(np.linalg.norm(new_xyz4[1:3, :], axis=0)[1:-1]) + 1
            max_xyz4 = xyz4[:, max_index].reshape(4, 1)
            max_xyz4_t = np.dot(T, max_xyz4)  # (4, 1)
            angle_x = angle(max_xyz4_t[1:3, 0], [1, 0])
            T = rotate3d_x(angle_x).dot(T)

        return T



"""Numpy related utils."""

from contextlib import contextmanager
import math
from typing import Any, Tuple

import numpy as np
import numpy.typing as npt

__all__ = ["padding1d", "XYPair", "to_distribution", "numpy_printoptions", "numpy_err"]


def padding1d(
    n: int,
    v: npt.NDArray | None,
    padding_value: Any = 0,
    dtype: npt.DTypeLike | None = None,
) -> npt.NDArray:
    """Padding x to array of shape (n,).

    Parameters
    ----------
    n : int
        Size of vector.
    v : np.ndarray, optional
        Input vector.
    padding_value : any, default to `0`.
        If x.shape[0] is less than n, the rest will be filled with
        padding value.
    dtype : np.DTypeLike, optional
        Data type of array. If specify, cast x to dtype, else dtype of
        x will used, otherwise defaults to `~numpy.float32`.
    """

    dtype = dtype or (v is not None and v.dtype) or np.float32
    v = np.zeros(n, dtype=dtype) if v is None else v
    v = v.astype(dtype) if v.dtype != dtype else v
    assert v.ndim == 1

    if v.shape[0] >= n:
        return v[:n]

    padding = np.full(n - v.shape[0], padding_value, dtype=dtype)
    return np.concatenate([v, padding])


XYPair = Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]


def to_distribution(
    values: npt.NDArray,
    step: float,
    vmin: float = 0,
    vmax: float | None = None,
    norm: bool = False,
) -> XYPair:
    indices = np.floor(values / step - vmin).astype(np.int32)
    size_min = np.max(indices).item() + 1
    size = math.ceil((vmax - vmin) / step) + 1 if vmax is not None else size_min
    y = np.zeros(max(size, size_min), dtype=np.float32)
    for i in indices:
        y[i] = y[i] + 1

    if norm:
        y /= values.shape[0]

    x = vmin + step * np.arange(size, dtype=np.float32)
    return x, y[:size]


@contextmanager
def numpy_printoptions(*args, **kwargs):
    original_options = np.get_printoptions()
    np.set_printoptions(*args, **kwargs)

    try:
        yield
    finally:
        np.set_printoptions(**original_options)


@contextmanager
def numpy_err(*args, **kwargs):
    old_settings = np.seterr(*args, **kwargs)

    try:
        yield
    finally:
        np.seterr(**old_settings)



"""SWC format."""

import warnings
from typing import Any, Dict, Iterable, List, Literal, Tuple, TypeVar, cast, overload

import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.sparse as sp

__all__ = ["swc_cols", "eswc_cols", "read_swc", "SWCLike", "SWCTypeVar"]

swc_cols: List[Tuple[str, npt.DTypeLike]] = [
    ("id", np.int32),
    ("type", np.int32),
    ("x", np.float32),
    ("y", np.float32),
    ("z", np.float32),
    ("r", np.float32),
    ("pid", np.int32),
]

eswc_cols: List[Tuple[str, npt.DTypeLike]] = [
    ("level", np.int32),
    ("mode", np.int32),
    ("timestamp", np.int32),
    ("teraflyindex", np.int32),
    ("feature_value", np.int32),
]


class SWCLike:
    """Abstract class that including swc infomation."""

    source: str = ""

    def __len__(self) -> int:
        return self.number_of_nodes()

    def id(self) -> npt.NDArray[np.int32]:  # pylint: disable=invalid-name
        """Get the ids of shape (n_sample,)."""
        return self.get_ndata("id")

    def type(self) -> npt.NDArray[np.int32]:
        """Get the types of shape (n_sample,)."""
        return self.get_ndata("type")

    def x(self) -> npt.NDArray[np.float32]:
        """Get the x coordinates of shape (n_sample,)."""
        return self.get_ndata("x")

    def y(self) -> npt.NDArray[np.float32]:
        """Get the y coordinates of shape (n_sample,)."""
        return self.get_ndata("y")

    def z(self) -> npt.NDArray[np.float32]:
        """Get the z coordinates of shape (n_sample,)."""
        return self.get_ndata("z")

    def r(self) -> npt.NDArray[np.float32]:
        """Get the radius of shape (n_sample,)."""
        return self.get_ndata("r")

    def pid(self) -> npt.NDArray[np.int32]:
        """Get the ids of parent of shape (n_sample,)."""
        return self.get_ndata("pid")

    def xyz(self) -> npt.NDArray[np.float32]:
        """Get the coordinates of shape(n_sample, 3)."""
        return np.stack([self.x(), self.y(), self.z()], axis=1)

    def xyzr(self) -> npt.NDArray[np.float32]:
        """Get the coordinates and radius array of shape(n_sample, 4)."""
        return np.stack([self.x(), self.y(), self.z(), self.r()], axis=1)

    def keys(self) -> Iterable[str]:
        raise NotImplementedError()

    def get_ndata(self, key: str) -> npt.NDArray[Any]:
        raise NotImplementedError()

    def get_adjacency_matrix(self) -> sp.coo_matrix:
        n_nodes = len(self)
        row, col = self.pid()[1:], self.id()[1:]  # ignore root
        triad = (np.ones_like(row), (row, col))
        return sp.coo_matrix(triad, shape=(n_nodes, n_nodes), dtype=np.int32)

    def number_of_nodes(self) -> int:
        """Get the number of nodes."""
        return self.id().shape[0]

    def number_of_edges(self) -> int:
        """Get the number of edges."""
        return self.number_of_nodes() - 1  # for tree structure: n = e + 1

    # fmt: off
    @overload
    def to_swc(self, swc_path: str, *, extra_cols: List[str] | None = ..., id_offset: int = ...) -> None: ...
    @overload
    def to_swc(self, *, extra_cols: List[str] | None = ..., id_offset: int = ...) -> str: ...
    # fmt: on
    def to_swc(
        self,
        swc_path: str | None = None,
        *,
        extra_cols: List[str] | None = None,
        id_offset: int = 1,
    ) -> str | None:
        """Write swc file."""
        it = self._to_swc(extra_cols=extra_cols, id_offset=id_offset)
        if swc_path is None:
            return "".join(it)

        with open(swc_path, "w", encoding="utf-8") as f:
            f.writelines(it)

    def _to_swc(self, extra_cols: List[str] | None, id_offset: int) -> Iterable[str]:
        def get_v(name: str, idx: int) -> str:
            vs = self.get_ndata(name)
            v = vs[idx]
            if np.issubdtype(vs.dtype, np.floating):
                return f"{v:.4f}"

            if name == "id" or (name == "pid" and v != -1):
                v += id_offset

            return str(v)

        names = [name for name, _ in swc_cols]
        if extra_cols is not None:
            names.extend(extra_cols)

        yield f"# source: {self.source if self.source else 'Unknown'}\n"
        yield f"# {' '.join(names)}\n"
        for idx in self.id():
            yield " ".join(get_v(name, idx) for name in names) + "\n"

    # fmt: off
    @overload
    def to_eswc(self, swc_path: str, **kwargs) -> None: ...
    @overload
    def to_eswc(self, **kwargs) -> str: ...
    # fmt: on
    def to_eswc(self, swc_path: str | None = None, **kwargs) -> str | None:
        kwargs.setdefault("swc_path", swc_path)
        kwargs.setdefault("extra_cols", [])
        kwargs["extra_cols"].extend(k for k, t in eswc_cols)
        return self.to_swc(**kwargs)


class DictSWC(SWCLike):
    ndata: Dict[str, npt.NDArray]

    def __init__(self, **kwargs: npt.NDArray):
        self.ndata = kwargs

    def keys(self) -> Iterable[str]:
        return self.ndata.keys()

    def get_ndata(self, key: str) -> npt.NDArray[Any]:
        return self.ndata[key]


SWCTypeVar = TypeVar("SWCTypeVar", bound=SWCLike)


def read_swc(
    swc_file: str,
    extra_cols: List[str] | None = None,
    fix_roots: Literal["somas", "nearest", False] = False,
    sort: bool = False,
    reset_index: bool = True,
) -> pd.DataFrame:
    """Read swc file.

    Parameters
    ----------
    swc_file : str
        Path of swc file, the id should be consecutively incremented.
    extra_cols : List[str], optional
        Read more cols in swc file.
    fix_roots : `somas`|`nearest`|False, default `False`
        Fix multiple roots.
    sort : bool, default `False`
        Sort the indices of neuron tree, the index for parent are
        always less than children.
    reset_index : bool, default `True`
        Reset node index to start with zero, DO NOT set to false if
        you are not sure what will happend.
    """

    names = [k for k, v in swc_cols]
    if extra_cols:
        names.extend(extra_cols)

    df = pd.read_csv(
        swc_file,
        sep=" ",
        comment="#",
        names=names,
        dtype=cast(Any, dict(swc_cols)),
        index_col=False,
    )

    # fix swc
    if fix_roots is not False and np.count_nonzero(df["pid"] == -1) > 1:
        match fix_roots:
            case "somas":
                swc_fix_roots_by_mark_somas(df)
            case "nearest":
                swc_fix_roots_by_link_nearest(df)
            case _:
                raise ValueError(f"unknown fix type: {fix_roots}")

    if sort:
        swc_sort(df)
    elif reset_index:
        swc_reset_index(df)

    # check swc
    if not swc_check_single_root(df):
        warnings.warn(f"core: not signle root, swc: {swc_file}")

    if (df["pid"] == -1).argmax() != 0:
        warnings.warn(f"core: root is not the first node, swc: {swc_file}")

    return df


def swc_fix_roots_by_mark_somas(
    df: pd.DataFrame, update_type: int | Literal[False] = 1
) -> None:
    """Merge multiple roots in swc.

    The first root are reserved and others are linked to it.
    """
    roots = df["pid"] == -1
    root_loc = roots.argmax()
    root_id = df.loc[root_loc, "id"]
    df["pid"] = np.where(df["pid"] != -1, df["pid"], root_id)
    if update_type is not False:
        df["type"] = np.where(df["pid"] != -1, df["type"], update_type)
    df.loc[root_loc, "pid"] = -1


def swc_fix_roots_by_link_nearest(df: pd.DataFrame) -> None:
    """Merge multiple roots in swc.

    The first root are reserved, and the others was.
    """
    dsu = _get_dsu(df)
    roots = df[df["pid"] == -1].iterrows()
    next(roots)  # type: ignore # skip the first one
    for i, row in roots:
        vs = df[["x", "y", "z"]] - row[["x", "y", "z"]]
        dis = np.linalg.norm(vs.to_numpy(), axis=1)
        subtree = dsu == dsu[i]  # type: ignore
        dis = np.where(subtree, np.Infinity, dis)  # avoid link to same tree
        dsu = np.where(subtree, dsu[dis.argmin()], dsu)  # merge set
        df.loc[i, "pid"] = df["id"][dis.argmin()]  # type: ignore


def swc_sort(df: pd.DataFrame):
    """Sort the indices of neuron tree.

    The index for parent are always less than children.
    """
    ids, pids = df["id"].to_numpy(), df["pid"].to_numpy()
    indices, new_ids, new_pids = swc_sort_impl(ids, pids)
    for col in df.columns:
        df[col] = df[col][indices].to_numpy()

    df["id"], df["pid"] = new_ids, new_pids


def swc_sort_impl(
    old_ids: npt.NDArray[np.int32], old_pids: npt.NDArray[np.int32]
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.int32], npt.NDArray[np.int32]]:
    """Sort the indices of neuron tree."""
    assert np.count_nonzero(old_pids == -1) == 1, "should be single root"

    id_map = np.full_like(old_ids, fill_value=-3)  # new_id to old_id
    new_pids = np.full_like(old_ids, fill_value=-3)
    new_id = 0
    s = [(old_ids[(old_pids == -1).argmax()], -1)]
    while len(s) != 0:
        old_id, new_pid = s.pop()
        id_map[new_id] = old_id
        new_pids[new_id] = new_pid
        s.extend((j, new_id) for j in old_ids[old_pids == old_id])  # (old_id, new_pid)
        new_id = new_id + 1

    id2idx = dict(zip(old_ids, range(len(old_ids))))  # old_id to old_idx
    indices = np.array([id2idx[i] for i in id_map], dtype=np.int32)  # new_id to old_idx
    new_ids = np.arange(len(new_pids))
    return indices, new_ids, new_pids


def swc_reset_index(df: pd.DataFrame) -> None:
    """Reset node index to start with zero."""
    roots = df["pid"] == -1
    root_loc = roots.argmax()
    root_id = df.loc[root_loc, "id"]
    df["id"] = df["id"] - root_id
    df["pid"] = df["pid"] - root_id
    df.loc[root_loc, "pid"] = -1


def swc_check_single_root(df: pd.DataFrame) -> bool:
    """Check is it only one root."""
    return len(np.unique(_get_dsu(df))) == 1


def _get_dsu(df: pd.DataFrame) -> npt.NDArray[np.int32]:
    dsu = np.where(df["pid"] == -1, df["id"], df["pid"])  # Disjoint Set Union

    id2idx = dict(zip(df["id"], range(len(df))))
    dsu = np.array([id2idx[i] for i in dsu], dtype=np.int32)

    while True:
        flag = True
        for i, p in enumerate(dsu):
            if dsu[i] != dsu[p]:
                dsu[i] = dsu[p]
                flag = False

        if flag:
            break

    return dsu



"""Sholl analysis."""

import math
from typing import Literal, Tuple

import numpy as np
import numpy.typing as npt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ..core import Tree
from ..utils import draw_circles, get_fig_ax

__all__ = ["Sholl"]


class Sholl:
    """Sholl analysis.

    .. [1] Dendritic organization in the neurons of the visual and motor
       cortices of the cat J. Anat., 87 (1953), pp. 387-406
    """

    count: npt.NDArray[np.int32]
    step: float

    def __init__(self, tree: Tree, step: float = 1) -> None:
        xyz = tree.get_segments().xyz() - tree.soma().xyz()  # shift
        radius = np.linalg.norm(xyz, axis=2)
        max_radius = math.ceil(np.max(radius[:, 1]))
        count = [
            np.count_nonzero(np.logical_and(radius[:, 0] <= i, radius[:, 1] > i))
            for i in np.arange(0, max_radius, step)
        ]

        self.count = np.array(count, dtype=np.int32)
        self.step = step

    def __getitem__(self, idx: int) -> int:
        return self.count[idx] if 0 <= idx < len(self.count) else 0

    def get_count(self) -> npt.NDArray[np.int32]:
        return self.count.copy()

    def get_distribution(self) -> Tuple[npt.NDArray[np.float32], npt.NDArray[np.int32]]:
        y = self.get_count()
        x = self.step * np.arange(len(y))
        return x, y

    def avg(self) -> float:
        return np.average(self.get_count()).item()

    def std(self) -> float:
        return np.std(self.get_count()).item()

    def sum(self) -> int:
        return np.sum(self.get_count()).item()

    def plot(
        self,
        plot_type: Literal["bar", "linechart", "circles"] = "linechart",
        fig: Figure | None = None,
        ax: Axes | None = None,
        **kwargs,
    ) -> Tuple[Figure, Axes]:
        fig, ax = get_fig_ax(fig, ax)
        x, y = self.get_distribution()

        match plot_type:
            case "bar":
                kwargs.setdefault("width", self.step)
                ax.bar(x, y, **kwargs)
            case "linechart":
                ax.plot(x, y, **kwargs)
            case "circles":
                kwargs.setdefault("y_min", 0)
                draw_circles(fig, ax, x, y, **kwargs)
            case _:
                raise ValueError(f"unsupported plot type: {plot_type}")

        return fig, ax



"""Analysis for neuron trees."""

from .branch_features import *
from .feature_extractor import *
from .node_features import *
from .path_features import *
from .sholl import *
from .visualization import *



"""Download helpers."""

import itertools
import logging
import multiprocessing
import os
from urllib.parse import urljoin

import bs4
import urllib3
import urllib3.exceptions

__all__ = ["download_file", "fetch_page", "clone_index_page"]


def download_file(dist: str, url: str) -> None:
    """Download a file."""
    conn = urllib3.connection_from_url(url)
    r = conn.request("GET", url)

    dirname = os.path.dirname(dist)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(dist, "wb") as file:
        file.write(r.data)


def fetch_page(url: str) -> bs4.BeautifulSoup:
    """Get page."""
    conn = urllib3.connection_from_url(url)
    r = conn.request("GET", url)
    data = r.data.decode("utf-8")
    return bs4.BeautifulSoup(data)


def clone_index_page(
    index_url: str, dist_dir: str, override: bool = False, multiprocess: int = 4
) -> None:
    """Download directory from index page.

    E.g: `https://download.brainimagelibrary.org/biccn/zeng/luo/fMOST/cells/`

    Parameters
    ----------
    index_url : str
        URL of index page.
    dist_dir : str
        Directory of dist.
    override : bool, default `False`
        Override existing file, skip file if `False`.
    multiprocess : int, default `4`
        How many process are available for download.
    """

    files = get_urls_in_index_page(index_url)
    logging.info("downloader: search `{}`, found {} files.", index_url, len(files))

    def download(url: str) -> None:
        filepath = url.removeprefix(index_url)
        dist = os.path.join(dist_dir, filepath)
        if os.path.exists(filepath):
            if not override:
                logging.info("downloader: file `{}` exits, skiped.", dist)
                return

            logging.info("downloader: file `{}` exits, deleted.", dist)
            os.remove(filepath)

        try:
            logging.info("downloader: downloading `{}` to `{}`", url, dist)
            download_file(filepath, url)
            logging.info("downloader: download `{}` to `{}`", url, dist)
        except urllib3.exceptions.HTTPError as ex:
            logging.info("downloader: fails to download `{}`, except `{}`", url, ex)

    with multiprocessing.Pool(multiprocess) as p:
        p.map(download, files)


def get_urls_in_index_page(url: str) -> list[str]:
    """Get all file links by dfs."""
    soup = fetch_page(url)
    links = [el.attrs["href"] for el in soup.find_all("a")]
    files = [urljoin(url, a) for a in links if not a.endswith("/")]
    dirs = [urljoin(url, a) for a in links if a != "../" and a.endswith("/")]
    files.extend(itertools.chain(*[get_urls_in_index_page(dir) for dir in dirs]))
    return files



"""Depth distribution of tree."""


from functools import cached_property
from typing import List

import numpy as np
import numpy.typing as npt

from ..core import BranchTree, Tree
from ..utils import to_distribution, XYPair

__all__ = ["NodeFeatures"]


class NodeFeatures:
    """Calc node feature of tree."""

    tree: Tree

    def __init__(self, tree: Tree) -> None:
        self.tree = tree

    def get_radial_distance(self) -> npt.NDArray[np.float32]:
        """Get the end-to-end straight-line distance to soma.

        Returns
        -------
        radial_distance : npt.NDArray[np.float32]
            Array of shape (N,), while N is the number of nodes.
        """
        xyz = self.tree.xyz() - self.tree.soma().xyz()
        radial_distance = np.linalg.norm(xyz, axis=1)
        return radial_distance

    def get_radial_distance_distribution(
        self,
        step: float = 1,
        /,
        filter_bifurcation: bool = False,
        filter_tip: bool = False,
        filter_other: bool = True,
    ) -> XYPair:
        """Get radial distance distribution of tree.

        Parameters
        ----------
        filter_bifurcation : bool, default `False`
            Filter bifurcation nodes.
        filter_tip : bool, default `False`
            Filter tip nodes.
        filter_other : bool, default `False`
            Filter nodes that are not bifurcations or tips.

        Returns
        -------
        radial_distance : npt.NDArray[np.float32]
            Array of shape (N,), while N is the number of nodes.
        """
        return self._to_distribution(
            self.get_radial_distance(),
            step,
            filter_bifurcation=filter_bifurcation,
            filter_tip=filter_tip,
            filter_other=filter_other,
        )

    def get_branch_order(self) -> npt.NDArray[np.int32]:
        """Get branch order of tree.

        Bifurcation order is the number of bifurcations between current position
        and the root.

        Returns
        -------
        order : npt.NDArray[np.int32]
            Array of shape (N,), while N is the number of nodes.
        """
        order = np.zeros_like(self._branch_tree.id(), dtype=np.int32)

        def assign_depth(n: Tree.Node, pre_depth: int | None) -> int:
            cur_order = pre_depth + 1 if pre_depth is not None else 0
            order[n.id] = cur_order
            return cur_order

        self._branch_tree.traverse(enter=assign_depth)
        return order

    def get_branch_order_distribution(
        self,
        step: int = 1,
        /,
        filter_bifurcation: bool = False,
        filter_tip: bool = False,
        filter_other: bool = True,
    ) -> XYPair:
        """Get branch order distribution of tree.

        Parameters
        ----------
        filter_bifurcation : bool, default `False`
            Filter bifurcation nodes.
        filter_tip : bool, default `False`
            Filter tip nodes.
        filter_other : bool, default `False`
            Filter nodes that are not bifurcations or tips.
        """
        return self._to_distribution(
            self.get_branch_order(),
            step,
            filter_bifurcation=filter_bifurcation,
            filter_tip=filter_tip,
            filter_other=filter_other,
        )

    def _to_distribution(
        self,
        x: npt.NDArray,
        step: float,
        /,
        filter_bifurcation: bool,
        filter_tip: bool,
        filter_other: bool,
    ) -> XYPair:
        if filter_bifurcation:
            x[self._bifurcations] = -1

        if filter_tip:
            x[self._tips] = -1

        if filter_other:
            x[self._other] = -1

        x = x[x != -1]
        return to_distribution(x, step)

    @cached_property
    def _branch_tree(self) -> BranchTree:
        return BranchTree.from_tree(self.tree)

    @cached_property
    def _bifurcations(self) -> List[int]:
        return [n.id for n in self.tree.get_bifurcations()]

    @cached_property
    def _tips(self) -> List[int]:
        return [n.id for n in self.tree.get_tips()]

    @cached_property
    def _other(self) -> npt.NDArray[np.int32]:
        other = self.tree.id()
        other = np.setdiff1d(other, self._bifurcations)
        other = np.setdiff1d(other, self._tips)
        return other



"""Tree utils."""

from typing import Callable, Dict, List, Tuple, TypeVar, cast, overload

import numpy as np
import numpy.typing as npt

from .swc import SWCLike, swc_sort_impl
from .tree import Tree

__all__ = ["REMOVE", "sort_tree", "to_sub_tree", "cut_tree", "propagate_remove"]

REMOVE = -2  # Set to `id`, mark as removed
T, K = TypeVar("T"), TypeVar("K")


def sort_tree(tree: Tree) -> Tree:
    """Sort the indices of neuron tree.

    See Also
    --------
    ~core.swc.swc_sort_tree
    """
    indices, new_ids, new_pids = swc_sort_impl(tree.id(), tree.pid())
    new_tree = tree.copy()
    new_tree.ndata = {k: tree.ndata[k][indices] for k in tree.ndata}
    new_tree.ndata["id"] = new_ids
    new_tree.ndata["pid"] = new_pids
    return new_tree


def to_sub_tree(
    swc_like: SWCLike, sub_id: npt.ArrayLike, sub_pid: npt.ArrayLike
) -> Tuple[Tree, Dict[int, int]]:
    """Create sub tree from origin tree.

    You can directly mark the node for removal, and we will
    automatically remove it, but if the node you remove is not a leaf
    node, you need to use `propagate_remove` to remove all children.
    """

    sub_id = np.array(sub_id, dtype=np.int32)
    sub_pid = np.array(sub_pid, dtype=np.int32)

    # remove nodes
    keeped_id = cast(npt.NDArray[np.bool_], sub_id != REMOVE)
    sub_id, sub_pid = sub_id[keeped_id], sub_pid[keeped_id]

    n_nodes = sub_id.shape[0]

    id_map = {idx: i for i, idx in enumerate(sub_id)}
    new_pid = [id_map[i] if i != -1 else -1 for i in sub_pid]

    ndata = {k: swc_like.get_ndata(k)[sub_id].copy() for k in swc_like.keys()}
    ndata.update(
        id=np.arange(0, n_nodes),
        pid=np.array(new_pid, dtype=np.int32),
    )

    new_tree = Tree(n_nodes, **ndata)
    new_tree.source = swc_like.source
    return new_tree, id_map


# fmt:off
@overload
def cut_tree(tree: Tree, *, enter: Callable[[Tree.Node, T | None], Tuple[T, bool]]) -> Tree: ...
@overload
def cut_tree(tree: Tree, *, leave: Callable[[Tree.Node, list[K]], Tuple[K, bool]]) -> Tree: ...
# fmt:on


def cut_tree(tree: Tree, *, enter=None, leave=None):
    """Cut tree.

    Returning a `True` can delete the current node and all its
    children.
    """

    idx, pid = tree.id().copy(), tree.pid().copy()
    if enter:

        def _enter(n: Tree.Node, parent: Tuple[T, bool] | None) -> Tuple[T, bool]:
            if parent is not None and parent[1]:
                idx[n.id] = REMOVE
                return parent

            res, remove = enter(n, parent[0] if parent else None)
            if remove:
                idx[n.id] = REMOVE

            return res, remove

        tree.traverse(enter=_enter)

    elif leave:

        def _leave(n: Tree.Node, children: List[K]) -> K:
            res, remove = leave(n, children)
            if remove:
                idx[n.id] = REMOVE

            return res

        tree.traverse(leave=_leave)
        propagate_remove(tree, idx)

    else:
        return tree.copy()

    new_tree, _ = to_sub_tree(tree, idx, pid)
    return new_tree


def propagate_remove(tree: Tree, idx: npt.NDArray[np.int32]) -> None:
    """Remove all children when parent is marked as removed."""

    def propagate(n: Tree.Node, remove_parent: bool | None) -> bool:
        remove = bool(remove_parent) or (idx[n.id] == REMOVE)
        if remove:
            idx[n.id] = REMOVE

        return remove

    tree.traverse(enter=propagate)



"""The segment is a branch with two nodes."""

from typing import Generic, Iterable, List, TypeVar

import numpy as np
import numpy.typing as npt

from .path import Path
from .swc import DictSWC, SWCTypeVar

__all__ = ["Segment", "Segments"]


class Segment(Path, Generic[SWCTypeVar]):
    r"""Segment attached to external object."""

    attach: SWCTypeVar
    idx: npt.NDArray[np.int32]

    def __init__(self, attach: SWCTypeVar, pid: int, idx: int) -> None:
        super().__init__(attach, np.array([pid, idx]))

    def keys(self) -> Iterable[str]:
        return self.attach.keys()

    def get_ndata(self, key: str) -> npt.NDArray:
        return self.attach.get_ndata(key)[self.idx]

    def detach(self) -> "Segment":
        """Detach from current attached object."""
        attact = DictSWC(**{k: self[k] for k in self.keys()})
        return Segment(attact, self.idx[0], self.idx[1])


SegmentT = TypeVar("SegmentT", bound=Segment)


class Segments(List[SegmentT]):
    r"""Segments contains a set of segments."""

    def __init__(self, segments: Iterable[SegmentT]) -> None:
        super().__init__(segments)

    def id(self) -> npt.NDArray[np.int32]:  # pylint: disable=invalid-name
        """Get the ids of shape (n_sample, 2)."""
        return self.get_ndata("id")

    def type(self) -> npt.NDArray[np.int32]:
        """Get the types of shape (n_sample, 2)."""
        return self.get_ndata("type")

    def x(self) -> npt.NDArray[np.float32]:
        """Get the x coordinates of shape (n_sample, 2)."""
        return self.get_ndata("x")

    def y(self) -> npt.NDArray[np.float32]:
        """Get the y coordinates of shape (n_sample, 2)."""
        return self.get_ndata("y")

    def z(self) -> npt.NDArray[np.float32]:
        """Get the z coordinates of shape (n_sample, 2)."""
        return self.get_ndata("z")

    def r(self) -> npt.NDArray[np.float32]:
        """Get the radius of shape (n_sample, 2)."""
        return self.get_ndata("r")

    def pid(self) -> npt.NDArray[np.int32]:
        """Get the ids of parent of shape (n_sample, 2)."""
        return self.get_ndata("pid")

    def xyz(self):
        """Get the coordinates of shape (n_sample, 2, 3)."""
        return np.stack([self.x(), self.y(), self.z()], axis=2)

    def xyzr(self):
        """Get the xyzr array of shape (n_sample, 2, 4)."""
        return np.stack([self.x(), self.y(), self.z(), self.r()], axis=2)

    def get_ndata(self, key: str) -> npt.NDArray:
        """Get ndata of shape (n_sample, 2).

        The order of axis 1 is (parent, current node).
        """
        return np.array([s.get_ndata(key) for s in self])



"""Create image stack from morphology."""

import re
import os
import time
from typing import Any, Iterable, List, Tuple, cast

import math
import numpy as np
import numpy.typing as npt
import tifffile
from tqdm import tqdm

from ..core import Tree, Population
from ..utils import SDF, SDFCompose, SDFRoundCone
from .base import Transform

__all__ = ["ToImageStack"]


class ToImageStack(Transform[Tree, npt.NDArray[np.uint8]]):
    r"""Transform tree to image stack."""

    resolution: npt.NDArray[np.float32]
    msaa: int
    z_per_iter: int = 1

    def __init__(
        self, resolution: float | npt.ArrayLike = 1, msaa: int = 8, z_per_iter: int = 1
    ) -> None:
        """Transform tree to image stack.

        Parameters
        ----------
        resolution : int | (x, y, z), default `(1, 1, 1)`
            Resolution of image stack.
        mass : int, default `8`
            Multi-sample anti-aliasing.
        """

        if isinstance(resolution, float):
            resolution = [resolution, resolution, resolution]

        self.resolution = (resolution := np.array(resolution, dtype=np.float32))
        assert tuple(resolution.shape) == (3,), "resolution shoule be vector of 3d."

        self.msaa = msaa
        self.z_per_iter = z_per_iter

    def __call__(self, x: Tree) -> npt.NDArray[np.uint8]:
        """Transform tree to image stack.

        Notes
        -----
        This method loads the entire image stack into memory, so it
        ONLY works for small image stacks, use
        :meth`transform_and_save` for big image stack.
        """
        return np.concatenate(list(self.transfrom(x, verbose=False)), axis=0)

    def __repr__(self) -> str:
        return (
            "ToImageStack"
            + f"-resolution-{'-'.join(self.resolution)}"
            + f"-mass-{self.msaa}"
            + f"-z-{self.z_per_iter}"
        )

    def transfrom(
        self, x: Tree, verbose: bool = True
    ) -> Iterable[npt.NDArray[np.uint8]]:
        if verbose:
            print("To image stack: " + x.source)
            time_start = time.time()

        sdf = self.get_sdf(x)

        xyz, r = x.xyz(), np.stack([x.r(), x.r(), x.r()], axis=1)  # TODO: perf
        coord_min = np.floor(np.min(xyz - r, axis=0))  # TODO: snap to grid
        coord_max = np.ceil(np.max(xyz + r, axis=0))
        grids, total = self.get_grids(coord_min, coord_max)

        if verbose:
            time_end = time.time()
            print("Prepare in: ", time_end - time_start, "s")  # type: ignore

        for grid in tqdm(grids, total=total) if verbose else grids:
            is_in = sdf.is_in(grid.reshape(-1, 3)).reshape(*grid.shape[:4])
            level = np.sum(is_in, axis=3, dtype=np.uint8)
            voxel = (255 / self.msaa * level).astype(np.uint8)
            for i in range(voxel.shape[2]):
                yield voxel[:, :, i]

    def transform_and_save(self, fname: str, x: Tree, verbose: bool = True) -> None:
        self.save_tif(fname, self.transfrom(x, verbose=verbose))

    def transform_population(
        self, population: Population | str, verbose: bool = True
    ) -> None:
        if isinstance(population, str):
            population = Population.from_swc(population)

        # TODO: multiprocess
        for tree in population:
            tif = re.sub(r".swc$", ".tif", tree.source)
            if not os.path.isfile(tif):
                self.transform_and_save(tif, tree, verbose=verbose)

    def get_grids(
        self, coord_min: npt.ArrayLike, coord_max: npt.ArrayLike
    ) -> Tuple[Iterable[npt.NDArray[np.float32]], int]:
        """Get point grid.

        Parameters
        ----------
        coord_min, coord_max: npt.ArrayLike
            Coordinates array of shape (3,).
        z_per_iter : int
            Yeild z per iter, raising this option speeds up processing,
            but consumes more memory.

        Returns
        -------
        grid : npt.NDArray[np.float32]
            Array of shape (nx, ny, z_per_iter, k, 3).
        """

        k = np.cbrt(self.msaa)
        coord_min, coord_max = np.array(coord_min), np.array(coord_max)
        assert tuple(coord_min.shape) == (3,), "coord_min shoule be vector of 3d."
        assert tuple(coord_max.shape) == (3,), "coord_max shoule be vector of 3d."

        point_grid = np.mgrid[
            coord_min[0] : coord_max[0] : self.resolution[0],
            coord_min[1] : coord_max[1] : self.resolution[1],
            coord_min[2] : coord_max[2] : self.resolution[2],
        ]  # (3, nx, ny, nz)
        point_grid = np.rollaxis(point_grid, 0, 4)  # (nx, ny, nz, 3)

        step = self.resolution / (k + 1)
        ends = self.resolution - step / 2
        inter_grid = np.mgrid[
            step[0] : ends[0] : step[0],
            step[1] : ends[1] : step[1],
            step[2] : ends[2] : step[2],
        ]  # (3, kx, ky, kz)
        inter_grid = np.rollaxis(inter_grid, 0, 4).reshape(-1, 3)  # (k, 3)

        grids = np.expand_dims(point_grid, 3).repeat(k**3, axis=3)
        grids = cast(Any, grids + inter_grid)

        return (
            grids[:, :, i : i + self.z_per_iter]
            for i in range(0, grids.shape[2], self.z_per_iter)
        ), math.ceil(grids.shape[2] / self.z_per_iter)

    def get_sdf(self, x: Tree) -> SDF:
        T = Tuple[Tree.Node, List[SDF], SDF | None]

        def collect(n: Tree.Node, pre: List[T]) -> T:
            if len(pre) == 0:
                return (n, [], None)

            if len(pre) == 1:
                child, sub_sdfs, last = pre[0]
                sub_sdfs.append(SDFRoundCone(n.xyz(), child.xyz(), n.r, child.r))
                return (n, sub_sdfs, last)

            sdfs: List[SDF] = []
            for child, sub_sdfs, last in pre:
                sub_sdfs.append(SDFRoundCone(n.xyz(), child.xyz(), n.r, child.r))
                sdfs.append(SDFCompose.compose(sub_sdfs))
                if last is not None:
                    sdfs.append(last)

            return (n, [], SDFCompose.compose(sdfs))

        _, sdfs, last = x.traverse(leave=collect)
        if len(sdfs) != 0:
            sdf = SDFCompose.compose(sdfs)
            if last is not None:
                sdf = SDFCompose.compose([sdf, last])
        elif last is not None:
            sdf = last
        else:
            raise ValueError("empty tree")

        return sdf

    @staticmethod
    def save_tif(
        fname: str,
        frames: Iterable[npt.NDArray[np.uint8]],
        resolution: Tuple[float, float] = (1, 1),
    ) -> None:
        with tifffile.TiffWriter(fname) as tif:
            for frame in frames:
                tif.write(
                    frame,
                    contiguous=True,
                    photometric="minisblack",
                    resolution=resolution,
                    metadata={
                        "unit": "um",
                        "axes": "ZXY",
                    },
                )



"""Branch tree is a simplified neuron tree."""

import itertools
from typing import Dict, List

from .branch import Branch
from .tree import Tree
from .tree_utils import to_sub_tree

__all__ = ["BranchTree"]


class BranchTree(Tree):
    """Branch tree keeps topology of tree.

    A branch tree that contains only soma, branch, and tip nodes.
    """

    branches: Dict[int, List[Branch]]

    def get_origin_branches(self) -> list[Branch]:
        """Get branches of original tree."""
        return list(itertools.chain(*self.branches.values()))

    def get_origin_node_branches(self, idx: int) -> List[Branch]:
        """Get branches of node of original tree."""
        return self.branches[idx]

    @staticmethod
    def from_tree(tree: Tree) -> "BranchTree":
        """Generating a branch tree from tree."""

        branches = tree.get_branches()

        sub_id = [br[-1].id for br in branches]
        sub_pid = [br[0].id for br in branches]
        # insert root
        sub_id.insert(0, 0)
        sub_pid.insert(0, -1)

        sub_tree, id_map = to_sub_tree(tree, sub_id, sub_pid)
        branch_tree = BranchTree(len(sub_tree), **sub_tree.ndata)
        branch_tree.source = tree.source  # TODO

        branch_tree.branches = {}
        for branch_raw in branches:
            idx = id_map[branch_raw[0].id]
            branch_tree.branches.setdefault(idx, [])
            branch_tree.branches[idx].append(branch_raw.detach())

        return branch_tree

    @staticmethod
    def from_swc(swc_file: str, **kwargs) -> "BranchTree":
        tree = super().from_swc(swc_file, **kwargs)
        return BranchTree.from_tree(tree)



"""Neuron tree."""

import itertools
import logging
import os
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
import numpy.typing as npt

from ..utils import padding1d
from .branch import Branch
from .node import Node
from .path import Path
from .segment import Segment, Segments
from .swc import SWCLike, eswc_cols, read_swc, swc_cols

__all__ = ["Tree"]

T, K = TypeVar("T"), TypeVar("K")


class Tree(SWCLike):
    """A neuron tree, which should be a binary tree in most cases."""

    class Node(Node["Tree"]):
        """Neural node."""

        def parent(self) -> Union["Tree.Node", None]:
            return Tree.Node(self.attach, self.pid) if self.pid != -1 else None

        def children(self) -> List["Tree.Node"]:
            return [Tree.Node(self.attach, idx) for idx in self.child_ids()]

        def is_soma(self) -> bool:
            return self.id == 0

        def is_tip(self) -> bool:
            return self.id not in self.attach.pid()

        def get_branch(self) -> "Tree.Branch":
            logging.info(
                "`tree.get_branch` has been renamed to `tree.branch` and will be removed in next version"
            )
            return self.branch()

        def branch(self) -> "Tree.Branch":
            nodes: List["Tree.Node"] = [self]
            while not (nodes[-1].is_soma() or nodes[-1].is_bifurcation()):
                nodes.append(cast(Tree.Node, nodes[-1].parent()))

            nodes.reverse()
            while not (nodes[-1].is_bifurcation() or nodes[-1].is_tip()):
                nodes.append(nodes[-1].children()[0])

            return Tree.Branch(self.attach, [n.id for n in nodes])

        def radial_distance(self) -> float:
            """The end-to-end straight-line distance to soma."""
            return self.distance(self.attach.soma())

    class Path(Path["Tree"]):
        """Neural path."""

    class Segment(Segment["Tree"]):
        """Neural segment."""

    class Branch(Branch["Tree"]):
        """Neural branch."""

    ndata: dict[str, npt.NDArray]

    def __init__(
        self,
        n_nodes: int,
        *,
        id: npt.NDArray[np.int32] | None = None,  # pylint: disable=redefined-builtin
        type: npt.NDArray[np.int32] | None = None,  # pylint: disable=redefined-builtin
        x: npt.NDArray[np.float32] | None = None,
        y: npt.NDArray[np.float32] | None = None,
        z: npt.NDArray[np.float32] | None = None,
        r: npt.NDArray[np.float32] | None = None,
        pid: npt.NDArray[np.int32] | None = None,
        **kwargs: npt.NDArray,
    ) -> None:
        id = np.arange(0, n_nodes, step=1, dtype=np.int32) if id is None else id
        pid = np.arange(-1, n_nodes - 1, step=1, dtype=np.int32) if pid is None else pid

        ndata = {
            "id": padding1d(n_nodes, id, dtype=np.int32),
            "type": padding1d(n_nodes, type, dtype=np.int32),
            "x": padding1d(n_nodes, x),
            "y": padding1d(n_nodes, y),
            "z": padding1d(n_nodes, z),
            "r": padding1d(n_nodes, r, padding_value=1),
            "pid": padding1d(n_nodes, pid, dtype=np.int32),
        }
        kwargs.update(ndata)
        self.ndata = kwargs

    def __iter__(self) -> Iterable[Node]:
        return (self[i] for i in range(len(self)))

    def __repr__(self) -> str:
        n_nodes, n_edges = self.number_of_nodes(), self.number_of_edges()
        return f"Neuron Tree with {n_nodes} nodes and {n_edges} edges"

    # fmt:off
    @overload
    def __getitem__(self, key: slice) -> List[Node]: ...
    @overload
    def __getitem__(self, key: int) -> Node: ...
    @overload
    def __getitem__(self, key: str) -> npt.NDArray: ...
    # fmt:on
    def __getitem__(self, key):
        if isinstance(key, slice):
            return [self.node(i) for i in range(*key.indices(len(self)))]

        if isinstance(key, int):
            length = len(self)
            if key < -length or key >= length:
                raise IndexError(f"The index ({key}) is out of range.")

            if key < 0:  # Handle negative indices
                key += length

            return self.node(key)

        if isinstance(key, str):
            return self.get_ndata(key)

        raise TypeError("Invalid argument type.")

    def keys(self) -> Iterable[str]:
        return self.ndata.keys()

    def node(self, idx: int) -> Node:
        return self.Node(self, idx)

    def soma(self) -> Node:
        return self.node(0)

    def get_ndata(self, key: str) -> npt.NDArray:
        return self.ndata[key]

    def get_bifurcations(self) -> List[Node]:
        """Get all node of bifurcations."""
        bifurcations: List[int] = []

        def collect_bifurcations(n: Tree.Node, children: List[None]) -> None:
            if len(children) > 1:
                bifurcations.append(n.id)

        self.traverse(leave=collect_bifurcations)
        return [self.node(i) for i in bifurcations]

    def get_tips(self) -> List[Node]:
        """Get all node of tips."""
        tip_ids = np.setdiff1d(self.id(), self.pid(), assume_unique=True)
        return [self.node(i) for i in tip_ids]

    def get_segments(self) -> Segments[Segment]:
        # pylint: disable-next=not-an-iterable
        return Segments([self.Segment(self, n.pid, n.id) for n in self[1:]])

    def get_branches(self) -> List[Branch]:
        Info = Tuple[List[Tree.Branch], List[int]]

        def collect_branches(node: "Tree.Node", pre: List[Info]) -> Info:
            if len(pre) == 1:
                branches, child = pre[0]
                child.append(node.id)
                return branches, child

            branches: List[Tree.Branch] = []

            for sub_branches, child in pre:
                child.append(node.id)
                child.reverse()
                sub_branches.append(Tree.Branch(self, np.array(child, dtype=np.int32)))
                sub_branches.reverse()
                branches.extend(sub_branches)

            return branches, [node.id]

        # pylint: disable-next=unpacking-non-sequence
        branches, _ = self.traverse(leave=collect_branches)
        return branches

    def get_paths(self) -> List[Path]:
        """Get all path from soma to tips."""
        path_dic: Dict[int, List[int]] = {}
        Paths = List[List[int]]

        def assign_path(n: Tree.Node, pre_path: List[int] | None) -> List[int]:
            path = [] if pre_path is None else pre_path.copy()
            path.append(n.id)
            path_dic[n.id] = path
            return path

        def collect_path(n: Tree.Node, children: List[Paths]) -> Paths:
            if len(children) == 0:
                return [path_dic[n.id]]

            return list(itertools.chain(*children))

        paths = self.traverse(enter=assign_path, leave=collect_path)
        # pylint: disable-next=not-an-iterable
        return [self.Path(self, idx) for idx in paths]

    # fmt: off
    @overload
    def traverse(self, *, enter: Callable[[Node, T | None], T], mode: Literal["dfs"] = ...) -> None: ...
    @overload
    def traverse(self, *, leave: Callable[[Node, list[K]], K], mode: Literal["dfs"] = ...) -> K: ...
    @overload
    def traverse(self, *,
        enter: Callable[[Node, T | None], T], leave: Callable[[Node, list[K]], K], mode: Literal["dfs"] = ...,
    ) -> K: ...
    # fmt: on

    def traverse(self, *, enter=None, leave=None, mode="dfs"):
        """Traverse each nodes.

        Parameters
        ----------
        enter : Callable[[Node, list[T]], T], optional
            The callback when entering each node, it accepts two parameters,
            the first parameter is the current node, the second parameter is
            the parent's information T, and the root node receives an None.
        leave : Callable[[Node, T | None], T], optional
            The callback when leaving each node. When leaving a node, subtree
            has already been traversed. Callback accepts two parameters, the
            first parameter is the current node, the second parameter is the
            children's information T, and the leaf node receives an empty list.
        """

        match mode:
            case "dfs":
                return self._traverse_dfs(enter=enter, leave=leave)
            case _:
                raise ValueError(f"unsupported mode: `{mode}`")

    def _traverse_dfs(self, *, enter=None, leave=None):
        """Traverse each nodes by dfs."""
        children_map = dict[int, list[int]]()
        for idx, pid in enumerate(self.pid()):
            children_map.setdefault(pid, [])
            children_map[pid].append(idx)

        # manual stack to avoid stack overflow in long projection
        stack: List[Tuple[int, bool]] = [(0, True)]  # (idx, first)
        params = {0: None}
        vals = {}

        while len(stack) != 0:
            idx, first = stack.pop()
            if first:
                pre = params.pop(idx)
                cur = enter(self[idx], pre) if enter is not None else None
                stack.append((idx, False))
                for child in children_map.get(idx, []):
                    stack.append((child, True))
                    params[child] = cur
            else:
                children = [vals.pop(i) for i in children_map.get(idx, [])]
                vals[idx] = leave(self[idx], children) if leave is not None else None

        return vals[0]

    def copy(self) -> "Tree":
        """Make a copy."""
        new_tree = Tree(len(self), **{k: v.copy() for k, v in self.ndata.items()})
        new_tree.source = self.source
        return new_tree

    def length(self) -> float:
        """Get length of tree."""
        return sum(s.length() for s in self.get_segments())

    @staticmethod
    def from_swc(swc_file: str, **kwargs) -> "Tree":
        """Read neuron tree from swc file.

        See Also
        --------
        ~swcgeom.read_swc
        """

        df = read_swc(swc_file, **kwargs)
        tree = Tree(df.shape[0], **{k: df[k].to_numpy() for k, v in swc_cols})
        tree.source = os.path.abspath(swc_file)
        return tree

    @staticmethod
    def from_eswc(swc_file: str, **kwargs) -> "Tree":
        """Read neuron tree from eswc file.

        See Also
        --------
        ~swcgeom.read_swc
        """

        kwargs.setdefault("extra_cols", [])
        kwargs["extra_cols"].extend(k for k, t in eswc_cols)
        return Tree.from_swc(swc_file, **kwargs)



"""Depth distribution of tree."""


from functools import cached_property
from typing import List

import numpy as np
import numpy.typing as npt

from ..core import Tree

__all__ = ["PathFeatures"]


class PathFeatures:
    """Path analysis of tree."""

    tree: Tree

    def __init__(self, tree: Tree) -> None:
        self.tree = tree

    def get_count(self) -> int:
        return len(self._paths)

    def get_length(self) -> npt.NDArray[np.float32]:
        """Get length of paths."""
        length = [path.length() for path in self._paths]
        return np.array(length, dtype=np.float32)

    def get_tortuosity(self) -> npt.NDArray[np.float32]:
        """Get tortuosity of path."""
        return np.array([path.tortuosity() for path in self._paths], dtype=np.float32)

    @cached_property
    def _paths(self) -> List[Tree.Path]:
        return self.tree.get_paths()



"""Painter utils."""

from typing import Dict, NamedTuple, Tuple

import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.axes import Axes
from matplotlib.collections import LineCollection
from matplotlib.colors import Normalize
from matplotlib.figure import Figure
from matplotlib.patches import Circle

import numpy as np
import numpy.typing as npt

from .transforms import (
    model_view_trasformation,
    orthographic_projection_simple,
    to_homogeneous,
    translate3d,
)

__all__ = [
    "Vector3D",
    "Camera",
    "palette",
    "draw_lines",
    "draw_xyz_axes",
    "draw_circles",
    "get_fig_ax",
]

Vector3D = Tuple[float, float, float]
Camera = NamedTuple("Camera", position=Vector3D, look_at=Vector3D, up=Vector3D)


class Palette:
    default: Dict[int, str]
    vaa3d: Dict[int, str]

    def __init__(self):
        default = [
            "#F596AA",  # momo,
            "#867835",  # kimirucha,
            "#E2943B",  # kuchiba,
            "#00896C",  # aotake,
            "#B9887D",  # mizugaki,
            "#2EA9DF",  # tsuyukusa,
            "#66327C",  # sumire,
            "#52433D",  # benikeshinezumi,
        ]
        self.default = dict(enumerate(default))

        vaa3d = [
            "#ffffff",  # white, 0-undefined
            "#141414",  # black, 1-soma
            "#c81400",  # red, 2-axon
            "#0014c8",  # blue, 3-dendrite
            "#c800c8",  # purple, 4-apical dendrite
            # the following is Hanchuan’s extended color. 090331
            "#00c8c8",  # cyan, 5
            "#dcc800",  # yellow, 6
            "#00c814",  # green, 7
            "#bc5e25",  # coffee, 8
            "#b4c878",  # asparagus, 9
            "#fa6478",  # salmon, 10
            "#78c8c8",  # ice, 11
            "#6478c8",  # orchid, 12
            # the following is Hanchuan’s further extended color. 111003
            "#ff80a8",  # 13
            "#80ffa8",  # 14
            "#80a8ff",  # 15
            "#a8ff80",  # 16
            "#ffa880",  # 17
            "#a880ff",  # 18
            "#000000",  # 19 # totally black. PHC, 2012-02-15
            # the following (20-275) is used for matlab heat map. 120209 by WYN
            "#000083",
        ]
        self.vaa3d = dict(enumerate(vaa3d))


palette = Palette()


def draw_lines(
    lines: npt.NDArray[np.floating], ax: Axes, camera: Camera, **kwargs
) -> LineCollection:
    """Draw lines.

    Parameters
    ----------
    lines : A collection of coords of lines
        Excepting a ndarray of shape (N, 2, 3), the axis-2 holds two points,
        and the axis-3 holds the coordinates (x, y, z).
    ax : ~matplotlib.axes.Axes
    camera : Camera
        Camera position.
    **kwargs : dict[str, Unknown]
        Forwarded to `~matplotlib.collections.LineCollection`.
    """

    starts, ends = lines[:, 0], lines[:, 1]
    starts, ends = to_homogeneous(starts, 1), to_homogeneous(ends, 1)

    # model/view transformation
    T = model_view_trasformation(*camera)
    T = translate3d(*camera[0]).dot(T)  # keep origin

    # projection transformation
    T = orthographic_projection_simple().dot(T)

    starts = np.dot(T, starts.T).T[:, 0:2]
    ends = np.dot(T, ends.T).T[:, 0:2]

    edges = np.stack([starts, ends], axis=1)
    collection = LineCollection(edges, **kwargs)  # type: ignore
    ax.add_collection(collection)  # type: ignore
    return collection


def draw_xyz_axes(
    ax: Axes, camera: Camera, position: Tuple[float, float] = (0.8, 0.10)
) -> None:
    x, y = position
    arrow_length = 0.05

    direction = np.array(
        [
            [1, 0, 0, 1],
            [0, 1, 0, 1],
            [0, 0, 1, 1],
            [0, 0, 0, 1],
        ]
    )
    direction = model_view_trasformation(*camera).dot(direction)

    for (dx, dy, dz, _), (text, color) in zip(
        direction[:3],
        [["x", "red"], ["y", "green"], ["z", "blue"]],
    ):
        if 1 - abs(dz) < 1e-5:
            continue

        ax.arrow(
            x,
            y,
            arrow_length * dx,
            arrow_length * dy,
            head_length=0.04,
            head_width=0.03,
            color=color,
            transform=ax.transAxes,
        )
        ax.text(
            x + (arrow_length + 0.06) * dx,
            y + (arrow_length + 0.06) * dy,
            text,
            color=color,
            transform=ax.transAxes,
        )


def draw_circles(
    fig: Figure,
    ax: Axes,
    x: npt.NDArray,
    y: npt.NDArray,
    /,
    y_min: float | None = None,
    y_max: float | None = None,
    cmap: str = "viridis",
) -> None:
    """Draw a sequential of circles."""
    y_min = y.min() if y_min is None else y_min
    y_max = y.max() if y_max is None else y_max
    norm = Normalize(y_min, y_max)

    color_map = cm.get_cmap(name=cmap)
    colors = color_map(norm(y))

    for xi, color in reversed(list(zip(x, colors))):
        circle = Circle((0, 0), xi, color=color)
        ax.add_patch(circle)

    sm = cm.ScalarMappable(cmap=color_map, norm=norm)
    sm.set_array([])
    fig.colorbar(sm)

    ax.set_aspect(1)
    ax.autoscale()


def get_fig_ax(
    fig: Figure | None = None, ax: Axes | None = None
) -> Tuple[Figure, Axes]:
    if fig is None:
        fig = plt.gcf() if ax is None else ax.get_figure()

    if ax is None:
        ax = fig.gca()

    return fig, ax



"""Nueron node."""

from typing import Any, Generic, Iterable

import numpy as np
import numpy.typing as npt

from .swc import DictSWC, SWCTypeVar

__all__ = ["Node"]


class Node(Generic[SWCTypeVar]):
    """Neural node."""

    attach: SWCTypeVar
    idx: int

    # fmt: off
    @property
    def id(self) -> int: return self["id"]
    @id.setter
    def id(self, v: int): self["id"] = v

    @property
    def type(self) -> int: return self["type"]
    @type.setter
    def type(self, v: int): self["type"] = v

    @property
    def x(self) -> float: return self["x"]
    @x.setter
    def x(self, v: float): self["x"] = v

    @property
    def y(self) -> float: return self["y"]
    @y.setter
    def y(self, v: float): self["y"] = v

    @property
    def z(self) -> float: return self["z"]
    @z.setter
    def z(self, v: float): self["z"] = v

    @property
    def r(self) -> float: return self["r"]
    @r.setter
    def r(self, v: float): self["r"] = v

    @property
    def pid(self) -> int: return self["pid"]
    @pid.setter
    def pid(self, v: int): self["pid"] = v
    # fmt: on

    def __init__(self, attach: SWCTypeVar, idx: int) -> None:
        super().__init__()
        self.attach = attach
        self.idx = idx

    def __getitem__(self, key: str) -> Any:
        return self.attach.get_ndata(key)[self.idx]

    def __setitem__(self, k: str, v: Any) -> None:
        self.attach.get_ndata(k)[self.idx] = v

    def __str__(self) -> str:
        return self.format_swc()

    def __repr__(self) -> str:
        return self.format_swc()

    def xyz(self) -> npt.NDArray[np.float32]:
        """Get the `x`, `y`, `z` of node, an array of shape (3,)"""
        return np.array([self.x, self.y, self.z], dtype=np.float32)

    def xyzr(self) -> npt.NDArray[np.float32]:
        """Get the `x`, `y`, `z`, `r` of node, an array of shape (4,)"""
        return np.array([self.x, self.y, self.z, self.r], dtype=np.float32)

    def keys(self) -> Iterable[str]:
        return self.attach.keys()

    def distance(self, b: "Node") -> float:
        """Get the distance of two nodes."""
        return np.linalg.norm(self.xyz() - b.xyz()).item()

    def format_swc(self) -> str:
        """Get the SWC format string."""
        x, y, z, r = [f"{f:.4f}" for f in [self.x, self.y, self.z, self.r]]
        items = [self.id, self.type, x, y, z, r, self.pid]
        return " ".join(map(str, items))

    def child_ids(self) -> npt.NDArray[np.int32]:
        return self.attach.id()[self.attach.pid() == self.id]

    def is_bifurcation(self) -> bool:
        return len(self.child_ids()) > 1

    def is_tip(self) -> bool:
        return len(self.child_ids()) == 0

    def detach(self) -> "Node":
        """Detach from current attached object."""
        attact = DictSWC(**{k: self[k] for k in self.keys()})
        return Node(attact, self.idx)



"""Branch Dataset."""

import os
import warnings
from typing import Generic, Iterable, TypeVar, cast

import torch
import torch.utils.data

from ...core import Branch
from ...transforms import Identity, Transform
from ...utils import numpy_err
from .tree_folder_dataset import TreeFolderDataset

__all__ = ["BranchDataset"]

T = TypeVar("T")
identity = Identity[Branch]()


class BranchDataset(torch.utils.data.Dataset, Generic[T]):
    """An easy way to get branches."""

    swc_dir: str
    save: str | None
    transform: Transform[Branch, T]

    branches: list[T]

    def __init__(
        self,
        swc_dir: str,
        save: str | bool = True,
        transform: Transform[Branch, T] = identity,
    ) -> None:
        """Create branch dataset.

        Parameters
        ----------
        swc_dir : str
            Path of SWC file directory.
        save : Union[str, bool], default `True`
            Save branch data to file if not False. If `True`, automatically
            generate file name.
        transfroms : Transfroms[Branch, T], optional
            Branch transformations.

        See Also
        --------
        ~swcgeom.transforms : module
            Preset transform set.
        """

        self.swc_dir = swc_dir
        self.transform = transform

        if isinstance(save, str):
            self.save = save
        elif save:
            self.save = os.path.join(swc_dir, self.get_filename())
        else:
            self.save = None

        if self.save and os.path.exists(self.save):
            self.branches = torch.load(self.save)
            return

        self.branches = self.get_branches()
        if self.save:
            torch.save(self.branches, self.save)

    def __getitem__(self, idx: int) -> T:
        """Get branch."""
        return self.branches[idx]

    def __len__(self) -> int:
        """Get length of branches."""
        return len(self.branches)

    def get_filename(self) -> str:
        """Get filename."""
        trans_name = f"-{self.transform}" if self.transform is not identity else ""
        return f"BranchDataset{trans_name}.pt"

    def get_branches(self) -> list[T]:
        """Get all branches."""
        branches = list[T]()

        with numpy_err(all="raise"):
            for tree in TreeFolderDataset(self.swc_dir):
                try:
                    brs = tree.get_branches()
                    brs = [self.transform(br) for br in brs] if self.transform else brs
                    branches.extend(cast(Iterable[T], brs))
                except Exception as ex:  # pylint: disable=broad-except
                    warnings.warn(
                        f"BranchDataset: skip swc '{tree.source}', got warning from numpy: {ex}"
                    )

        return branches



"""3D geometry transformations."""

from typing import Tuple

import numpy as np
import numpy.typing as npt

__all__ = [
    "angle",
    "scale3d",
    "translate3d",
    "rotate3d",
    "rotate3d_x",
    "rotate3d_y",
    "rotate3d_z",
    "to_homogeneous",
    "model_view_trasformation",
    "orthographic_projection_simple",
]

Vector3D = Tuple[float, float, float]


def angle(a: npt.ArrayLike, b: npt.ArrayLike) -> float:
    """Get the agnle of vectors.

    Returns
    -------
    angle : float
        Angle in radians.
    """
    a, b = np.array(a), np.array(b)
    costheta = np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))
    theta = np.arccos(costheta)
    return theta if np.cross(a, b) > 0 else -theta


def scale3d(sx: float, sy: float, sz: float) -> npt.NDArray[np.float32]:
    """Get the 3D scale transfomation matrix.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """
    return np.array(
        [
            [sx, 0, 0, 0],
            [0, sy, 0, 0],
            [0, 0, sz, 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float32,
    )


def translate3d(tx: float, ty: float, tz: float) -> npt.NDArray[np.float32]:
    """Get the 3D translate transfomation matrix.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """
    return np.array(
        [
            [1, 0, 0, tx],
            [0, 1, 0, ty],
            [0, 0, 1, tz],
            [0, 0, 0, 1],
        ],
        dtype=np.float32,
    )


def rotate3d(n: npt.ArrayLike, theta: float) -> npt.NDArray[np.float32]:
    r"""Get the 3D rotation transfomation matrix.

    Rotate v with axis n by an angle theta according to the right hand rule,
    follow rodrigues' rotaion formula.

    .. math::

        R(\mathbf{n}, \alpha) = \cos{\alpha} \cdot \mathbf{I}
        + (1-\cos{\alpha}) \cdot \mathbf{n} \cdot \mathbf{n^T}
        + \sin{\alpha} \cdot \mathbf{N} \\

        N  = \begin{pmatrix}
        0 & -n_z & n_y \\
        n_z & 0 & -n_x \\
        -n_y & n_x & 0
        \end{pmatrix}

    Parameters
    ----------
    n : ArrayLike
        Rotation axis.
    theta : float
        Rotation angle in radians.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """

    n = np.array(n)
    nx, ny, nz = n[0:3]
    # pylint: disable-next=invalid-name
    N = np.array(
        [
            [0, -nz, ny],
            [nz, 0, -nx],
            [-ny, nx, 0],
        ],
        dtype=np.float32,
    )

    return (
        np.cos(theta) * np.identity(4)
        + (1 - np.cos(theta)) * n * n[:, None]
        + np.sin(theta) * N
    )


def rotate3d_x(theta: float) -> npt.NDArray[np.float32]:
    """Get the 3D rotation transfomation matrix.

    Rotate 3D vector `v` with `x`-axis by an angle theta according to the right
    hand rule.

    Parameters
    ----------
    theta : float
        Rotation angle in radians.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """

    return np.array(
        [
            [1, 0, 0, 0],
            [0, np.cos(theta), -np.sin(theta), 0],
            [0, np.sin(theta), np.cos(theta), 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float32,
    )


def rotate3d_y(theta: float) -> npt.NDArray[np.float32]:
    """Get the 3D rotation transfomation matrix.

    Rotate 3D vector `v` with `y`-axis by an angle theta according to the right
    hand rule.

    Parameters
    ----------
    theta : float
        Rotation angle in radians.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """
    return np.array(
        [
            [np.cos(theta), 0, np.sin(theta), 0],
            [0, 1, 0, 0],
            [-np.sin(theta), 0, np.cos(theta), 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float32,
    )


def rotate3d_z(theta: float) -> npt.NDArray[np.float32]:
    """Get the 3D rotation transfomation matrix.

    Rotate 3D vector `v` with `z`-axis by an angle theta according to the right
    hand rule.

    Parameters
    ----------
    theta : float
        Rotation angle in radians.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """
    return np.array(
        [
            [np.cos(theta), -np.sin(theta), 0, 0],
            [np.sin(theta), np.cos(theta), 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float32,
    )


def to_homogeneous(xyz: npt.ArrayLike, w: float) -> npt.NDArray[np.float32]:
    """Fill xyz to homogeneous coordinates.

    Parameters
    ----------
    xyz : ArrayLike
        Coordinate of shape (N, 3)
    w : float
        w of homogeneous coordinate, 1 for dot, 0 for vector.

    Returns
    -------
    xyz4 : npt.NDArray[np.float32]
        Array of shape (N, 4)
    """
    xyz = np.array(xyz)
    if xyz.shape[1] == 4:
        return xyz

    filled = np.full((xyz.shape[0], 1), fill_value=w)
    xyz4 = np.concatenate([xyz, filled], axis=1)
    return xyz4


def model_view_trasformation(
    position: Vector3D, look_at: Vector3D, up: Vector3D
) -> npt.NDArray[np.float32]:
    r"""Play model/view transformation.

    Parameters
    ----------
    position: Tuple[float, float, float]
        Camera position \vec{e}.
    look_at: Tuple[float, float, float]
        Camera look-at \vec{g}.
    up: Tuple[float, float, float]
        Camera up direction \vec{t}.
    """

    e = np.array(position, dtype=np.float32)
    g = np.array(look_at, dtype=np.float32) / np.linalg.norm(look_at)
    t = np.array(up, dtype=np.float32) / np.linalg.norm(up)

    t_view = translate3d(*(-1 * e))
    r_view = np.array(
        [
            [*np.cross(g, t), 0],
            [*t, 0],
            [*(-1 * g), 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float32,
    )
    return np.dot(r_view, t_view)


def orthographic_projection_simple() -> npt.NDArray[np.float32]:
    """Simple orthographic projection by drop z-axis"""
    return np.array(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 0],  # drop z-axis
            [0, 0, 0, 0],  # drop w
        ],
        dtype=np.float32,
    )



"""Branch is a set of node points."""

from typing import Generic, Iterable

import numpy as np
import numpy.typing as npt
from typing_extensions import Self  # TODO: move to typing in python 3.11

from .path import Path
from .segment import Segment, Segments
from .swc import DictSWC, SWCTypeVar

__all__ = ["Branch"]


class Branch(Path, Generic[SWCTypeVar]):
    r"""Neural branch.

    Notes
    -----
    Only a part of data of branch nodes is valid, such as `x`, `y`, `z` and
    `r`, but the `id` and `pid` is usually invalid.
    """

    attach: SWCTypeVar
    idx: npt.NDArray[np.int32]

    class Segment(Segment["Branch"]):
        """Segment of branch."""

    def __repr__(self) -> str:
        return f"Neuron branch with {len(self)} nodes."

    def keys(self) -> Iterable[str]:
        return self.attach.keys()

    def get_ndata(self, key: str) -> npt.NDArray:
        return self.attach.get_ndata(key)[self.idx]

    def get_segments(self) -> Segments[Segment]:
        return Segments([self.Segment(self, n.pid, n.id) for n in self[1:]])

    def detach(self) -> "Branch":
        """Detach from current attached object."""
        attact = DictSWC(**{k: self[k] for k in self.keys()})
        return Branch(attact, self.idx.copy())

    @classmethod
    def from_xyzr(cls, xyzr: npt.NDArray[np.float32]) -> Self:
        r"""Create a branch from ~numpy.ndarray.

        Parameters
        ----------
        xyzr : npt.NDArray[np.float32]
            Collection of nodes. If shape (n, 4), both `x`, `y`, `z`, `r` of
            nodes is enabled. If shape (n, 3), only `x`, `y`, `z` is enabled
            and `r` will fill by 1.
        """
        assert xyzr.ndim == 2 and xyzr.shape[1] in (
            3,
            4,
        ), f"xyzr should be of shape (N, 3) or (N, 4), got {xyzr.shape}"

        n_nodes = xyzr.shape[0]
        if xyzr.shape[1] == 3:
            ones = np.ones([n_nodes, 1], dtype=np.float32)
            xyzr = np.concatenate([xyzr, ones], axis=1)

        idx = np.arange(0, n_nodes, step=1, dtype=np.int32)
        attact = DictSWC(
            id=idx,
            type=np.full((n_nodes), fill_value=3, dtype=np.int32),
            x=xyzr[:, 0],
            y=xyzr[:, 1],
            z=xyzr[:, 2],
            r=xyzr[:, 3],
            pid=np.arange(-1, n_nodes - 1, step=1, dtype=np.int32),
        )
        return cls(attact, idx)

    @classmethod
    def from_xyzr_batch(cls, xyzr_batch: npt.NDArray[np.float32]) -> list[Self]:
        r"""Create list of branch form ~numpy.ndarray.

        Parameters
        ----------
        xyzr : npt.NDArray[np.float32]
            Batch of collection of nodes. If shape (bs, n, 4), both `x`, `y`,
            `z`, `r` of nodes is enabled. If shape (bs, n, 3), only `x`, `y`,
            `z` is enabled and `r` will fill by 1.
        """

        assert xyzr_batch.ndim == 3
        assert xyzr_batch.shape[1] >= 3

        if xyzr_batch.shape[2] == 3:
            ones = np.ones(
                [xyzr_batch.shape[0], xyzr_batch.shape[1], 1], dtype=np.float32
            )
            xyzr_batch = np.concatenate([xyzr_batch, ones], axis=2)

        branches = list[Branch]()
        for xyzr in xyzr_batch:
            n_nodes = xyzr.shape[0]
            idx = np.arange(0, n_nodes, step=1, dtype=np.int32)
            attact = DictSWC(
                id=idx,
                type=np.full((n_nodes), fill_value=3, dtype=np.int32),
                x=xyzr[:, 0],
                y=xyzr[:, 1],
                z=xyzr[:, 2],
                r=xyzr[:, 3],
                pid=np.arange(-1, n_nodes - 1, step=1, dtype=np.int32),
            )
            branches.append(cls(attact, idx))

        return branches



"""A seires of common feature.

Notes
-----
For development, see method `_Features.get_evaluator`
to confirm the naming specification.
"""

from functools import cached_property
from typing import Any, Callable, Dict, List, Literal, Tuple, cast, overload

import numpy as np
import numpy.typing as npt
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ..core import Population, Tree
from ..utils import XYPair, get_fig_ax, padding1d, to_distribution
from .branch_features import BranchFeatures
from .node_features import NodeFeatures
from .path_features import PathFeatures
from .sholl import Sholl

__all__ = [
    "Feature",
    "FeatureExtractor",
    "PopulationFeatureExtractor",
    "extract_feature",
]

Feature = Literal[
    "length",
    "sholl",
    # node
    "node_radial_distance",
    "node_branch_order",
    # branch
    "branch_length",
    "branch_tortuosity",
    # path
    "path_length",
    "path_tortuosity",
]


class Features:
    """Tree features"""

    tree: Tree

    # Modules
    # fmt:off
    @cached_property
    def _node_features(self) -> NodeFeatures: return NodeFeatures(self.tree)
    @cached_property
    def _branch_features(self) -> BranchFeatures: return BranchFeatures(self.tree)
    @cached_property
    def _path_features(self) -> PathFeatures: return PathFeatures(self.tree)
    # fmt:on

    def __init__(self, tree: Tree) -> None:
        self.tree = tree

    def get(self, feature: Feature, **kwargs) -> npt.NDArray[np.float32]:
        evaluator = self.get_evaluator(feature)
        feat = evaluator(**kwargs).astype(np.float32)
        return feat

    def get_evaluator(self, feature: Feature) -> Callable[[], npt.NDArray]:
        if callable(calc := getattr(self, f"get_{feature}", None)):
            return calc  # custom features

        components = feature.split("_")
        if (module := getattr(self, f"_{components[0]}_features", None)) and callable(
            calc := getattr(module, f"get_{'_'.join(components[1:])}", None)
        ):
            return calc

        raise ValueError(f"Invalid feature: {feature}")

    def get_distribution(
        self, feature: Feature, step: float | None, **kwargs
    ) -> XYPair:
        if callable(method := getattr(self, f"get_{feature}_distribution", None)):
            if step is not None:
                kwargs.setdefault("step", step)
            return method(**kwargs)  # custom feature distribution

        feat = self.get(feature, **kwargs)
        step = cast(float, feat.max() / 100) if step is None else step
        x, y = to_distribution(feat, step)
        return x, y

    # Features

    def get_length(self, **kwargs) -> npt.NDArray[np.float32]:
        return np.array([self.tree.length(**kwargs)], dtype=np.float32)

    def get_sholl(self, **kwargs) -> npt.NDArray[np.int32]:
        return Sholl(self.tree, **kwargs).get_count()

    def get_sholl_distribution(self, **kwargs) -> XYPair:
        x, y = Sholl(self.tree, **kwargs).get_distribution()
        return x, y.astype(np.float32)


class FeatureExtractor:
    """Extract feature from tree."""

    features: Features

    def __init__(self, tree: Tree) -> None:
        self.features = Features(tree)

    # fmt:off
    @overload
    def get(self, feature: Feature, **kwargs) -> npt.NDArray[np.float32]: ...
    @overload
    def get(self, feature: List[Feature], **kwargs) -> List[npt.NDArray[np.float32]]: ...
    @overload
    def get(self, feature: Dict[Feature, Dict[str, Any]], **kwargs) -> Dict[str, npt.NDArray[np.float32]]: ...
    # fmt:on
    def get(self, feature, **kwargs):
        """Get feature of shape (L,)."""
        if isinstance(feature, dict):
            return {k: self.features.get(k, **v) for k, v in feature.items()}

        if isinstance(feature, list):
            return [self.features.get(k) for k in feature]

        return self.features.get(feature, **kwargs)

    # fmt:off
    @overload
    def get_distribution(self, feature: Feature, step: float = ..., **kwargs) -> XYPair: ...
    @overload
    def get_distribution(self, feature: List[Feature], step: float = ..., **kwargs) -> List[XYPair]: ...
    @overload
    def get_distribution(self, feature: Dict[Feature, Dict[str, Any]], step: float = ..., **kwargs) -> Dict[str, XYPair]: ...
    # fmt:on
    def get_distribution(self, feature, step: float | None = None, **kwargs):
        """Get feature distribution of shape (S,)."""
        if isinstance(feature, dict):
            return {
                k: self.features.get_distribution(k, step, **v)
                for k, v in feature.items()
            }

        if isinstance(feature, list):
            return [self.features.get_distribution(k, step) for k in feature]

        return self.features.get_distribution(feature, step, **kwargs)

    def plot_distribution(
        self,
        feature: Feature,
        fig: Figure | None = None,
        ax: Axes | None = None,
        **kwargs,
    ) -> Tuple[Figure, Axes]:
        # pylint: disable=unpacking-non-sequence
        x, y = self.get_distribution(feature, **kwargs)
        fig, ax = get_fig_ax(fig, ax)
        sns.lineplot(x=x, y=y, ax=ax)
        return fig, ax


class PopulationFeatureExtractor:
    """Extract feature from population."""

    population: Population

    @cached_property
    def _trees(self) -> List[Features]:
        return [Features(tree) for tree in self.population]

    def __init__(self, population: Population) -> None:
        self.population = population

    # fmt:off
    @overload
    def get(self, feature: Feature, **kwargs) -> List[npt.NDArray[np.float32]]: ...
    @overload
    def get(self, feature: Dict[Feature, Dict[str, Any]], **kwargs) -> Dict[str, List[npt.NDArray[np.float32]]]: ...
    # fmt:on
    def get(self, feature, **kwargs):
        """Get feature list of array of shape (N, L_i).

        Which N is the number of tree of population, L is length of
        nodes.
        """
        if isinstance(feature, dict):
            return {k: self._get(k, **v) for k, v in feature.items()}

        return self._get(feature, **kwargs)

    # fmt:off
    @overload
    def get_distribution(self, feature: Feature, step: float = ..., **kwargs) -> XYPair: ...
    @overload
    def get_distribution(self, feature: List[Feature], step: float = ..., **kwargs) -> List[XYPair]: ...
    @overload
    def get_distribution(self, feature: Dict[Feature, Dict[str, Any]], step: float = ..., **kwargs) -> Dict[str, XYPair]: ...
    # fmt:on
    def get_distribution(self, feature, step: float | None = None, **kwargs):
        """Get feature distribution of shape (N, S).

        Which N is the number of tree of population, S is size of
        distrtibution.

        Returns
        -------
        x : npt.NDArray[np.float32]
            Array of shape (S,).
        y : npt.NDArray[np.float32]
            Array of shape (N, S).
        """
        if isinstance(feature, dict):
            return {
                k: self._get_distribution(k, step=step, **v) for k, v in feature.items()
            }

        if isinstance(feature, list):
            return [self._get_distribution(k, step=step) for k in feature]

        return self._get_distribution(feature, step=step, **kwargs)

    def plot_distribution(
        self,
        feature: Feature,
        fig: Figure | None = None,
        ax: Axes | None = None,
        **kwargs,
    ) -> Tuple[Figure, Axes]:
        # pylint: disable=unpacking-non-sequence
        x, y = self.get_distribution(feature, **kwargs)
        x, y = np.tile(x, y.shape[0]), y.flatten()

        fig, ax = get_fig_ax(fig, ax)
        sns.lineplot(x=x, y=y, ax=ax)
        return fig, ax

    def _get(self, feature: Feature, **kwargs) -> List[npt.NDArray[np.float32]]:
        return [ex.get(feature, **kwargs) for ex in self._trees]

    def _get_distribution(self, feature: Feature, **kwargs) -> XYPair:
        assert len(self._trees) != 0

        x, ys = np.array([], dtype=np.float32), list[npt.NDArray[np.float32]]()
        for features in self._trees:
            xx, y = features.get_distribution(feature, **kwargs)
            x = xx if xx.shape[0] > x.shape[0] else x
            ys.append(y)

        max_len_y = max(y.shape[0] for y in ys)
        y = np.stack([padding1d(max_len_y, y, 0) for y in ys])
        return x, y


# fmt: off
@overload
def extract_feature(obj: Tree) -> FeatureExtractor: ...
@overload
def extract_feature(obj: Population) -> PopulationFeatureExtractor: ...
# fmt: on
def extract_feature(obj):
    if isinstance(obj, Tree):
        return FeatureExtractor(obj)

    if isinstance(obj, Population):
        return PopulationFeatureExtractor(obj)

    raise TypeError("Invalid argument type.")



"""Signed distance functions.

Refs: https://iquilezles.org/articles/distfunctions/
"""

import warnings
from typing import List, Tuple, Iterable

import numpy as np
import numpy.typing as npt

__all__ = ["SDF", "SDFCompose", "SDFRoundCone"]

# Axis-aligned bounding box, tuple of array of shape (3,)
AABB = Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]


class SDF:
    """Signed distance functions."""

    bounding_box: AABB | None = None

    def __call__(self, p: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        return self.distance(p)

    def distance(self, p: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        """Calc signed distance.

        Parmeters
        ---------
        p: ArrayLike
            Hit point p of shape (N, 3).

        Returns
        -------
        distance : npt.NDArray[np.float32]
            Distance array of shape (3,).
        """
        raise NotImplementedError()

    def is_in(self, p: npt.NDArray[np.float32]) -> npt.NDArray[np.bool_]:
        p = np.array(p, dtype=np.float32)
        assert p.ndim == 2 and p.shape[1] == 3, "p should be array of shape (N, 3)"

        in_box = self.is_in_bounding_box(p)
        flags = np.full((p.shape[0]), False, dtype=np.bool_)
        flags[in_box] = self.distance(p[in_box]) <= 0
        return flags

    def is_in_bounding_box(self, p: npt.NDArray[np.float32]) -> npt.NDArray[np.bool_]:
        """Is p in bounding box.

        Returns
        -------
        is_in : npt.NDArray[np.bool_]
            Array of shape (N,), if bounding box is `None`, `True` will
            be returned.
        """

        if self.bounding_box is None:
            return np.full((p.shape[0]), True, dtype=np.bool_)

        is_in = np.logical_and(
            np.all(p >= self.bounding_box[0], axis=1),
            np.all(p <= self.bounding_box[1], axis=1),
        )
        return is_in


class SDFCompose(SDF):
    """Compose multiple SDFs."""

    def __init__(self, sdfs: List[SDF]) -> None:
        assert len(sdfs) != 0, "must combine at least one SDF"

        if len(sdfs) == 1:
            warnings.warn("compose only one SDF, use SDFCompose.compose instead")

        self.sdfs = sdfs

        bounding_boxes = [sdf.bounding_box for sdf in sdfs if sdf.bounding_box]
        if len(bounding_boxes) == len(self.sdfs):
            self.bounding_box = (
                np.min(np.stack([box[0] for box in bounding_boxes]).T, axis=1),
                np.max(np.stack([box[1] for box in bounding_boxes]).T, axis=1),
            )

    def distance(self, p: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        return np.min([sdf(p) for sdf in self.sdfs], axis=0)

    def is_in(self, p: npt.NDArray[np.float32]) -> npt.NDArray[np.bool_]:
        in_box = self.is_in_bounding_box(p)
        is_in = np.stack([sdf.is_in(p[in_box]) for sdf in self.sdfs])
        flags = np.full_like(in_box, False, dtype=np.bool_)
        flags[in_box] = np.any(is_in, axis=0)
        return flags

    @staticmethod
    def compose(sdfs: Iterable[SDF]) -> SDF:
        sdfs = list(sdfs)
        return SDFCompose(sdfs) if len(sdfs) != 1 else sdfs[0]


class SDFRoundCone(SDF):
    """Round cone is made up of two balls and a cylinder."""

    def __init__(
        self, a: npt.ArrayLike, b: npt.ArrayLike, ra: float, rb: float
    ) -> None:
        """SDF of round cone.

        Parmeters
        ---------
        a, b : ArrayLike
            Coordinates of point A/B of shape (3,).
        ra, rb : float
            Radius of point A/B.
        """

        self.a = np.array(a, dtype=np.float32)
        self.b = np.array(b, dtype=np.float32)
        self.ra = ra
        self.rb = rb

        assert tuple(self.a.shape) == (3,), "a should be vector of 3d"
        assert tuple(self.b.shape) == (3,), "b should be vector of 3d"

        self.bounding_box = (
            np.min([self.a - self.ra, self.b - self.rb], axis=0).astype(np.float32),
            np.max([self.a + self.ra, self.b + self.rb], axis=0).astype(np.float32),
        )

    def distance(self, p: npt.ArrayLike) -> npt.NDArray[np.float32]:
        p = np.array(p, dtype=np.float32)
        assert p.ndim == 2 and p.shape[1] == 3, "p should be array of shape (N, 3)"

        a = self.a
        b = self.b
        ra = self.ra
        rb = self.rb

        # sampling independent computations (only depend on shape)
        ba = b - a
        l2 = np.dot(ba, ba)
        rr = ra - rb
        a2 = l2 - rr * rr
        il2 = 1.0 / l2

        # sampling dependant computations
        pa = p - a
        y = np.dot(pa, ba)
        z = y - l2
        x = pa * l2 - np.outer(y, ba)
        x2 = np.sum(x * x, axis=1)
        y2 = y * y * l2
        z2 = z * z * l2

        # single square root!
        k = np.sign(rr) * rr * rr * x2
        dis = (np.sqrt(x2 * a2 * il2) + y * rr) * il2 - ra

        lt = np.sign(z) * a2 * z2 > k
        dis[lt] = np.sqrt(x2[lt] + z2[lt]) * il2 - rb

        rt = np.sign(y) * a2 * y2 < k
        dis[rt] = np.sqrt(x2[rt] + y2[rt]) * il2 - ra

        return dis



"""Neuron population is a set of tree."""

import os
from typing import Any, Dict, Iterator, List, cast, overload

from .swc import eswc_cols
from .tree import Tree

__all__ = ["Population"]


class Population:
    """Neuron population."""

    source: str = ""
    swcs: List[str]
    trees: List[Tree | None]
    read_kwargs: Dict[str, Any]

    def __init__(self, swcs: list[str], lazy_loading=True) -> None:
        super().__init__()
        self.swcs = swcs
        self.trees = [None for _ in swcs]
        if not lazy_loading:
            self.load(slice(len(swcs)))

    # fmt:off
    @overload
    def __getitem__(self, key: slice) -> List[Tree]: ...
    @overload
    def __getitem__(self, key: int) -> Tree: ...
    # fmt:on
    def __getitem__(self, key):
        self.load(key)
        if isinstance(key, slice):
            return [cast(Tree, self.trees[i]) for i in range(*key.indices(len(self)))]

        if isinstance(key, int):
            return cast(Tree, self.trees[key])

        raise TypeError("Invalid argument type.")

    def __len__(self) -> int:
        return len(self.swcs)

    def __iter__(self) -> Iterator[Tree]:
        return (self[i] for i in range(self.__len__()))

    def __repr__(self) -> str:
        return f"Neuron population in '{self.source}'"

    # fmt:off
    @overload
    def load(self, key: slice) -> None: ...
    @overload
    def load(self, key: int) -> None: ...
    # fmt:on
    def load(self, key):
        if isinstance(key, slice):
            idx = range(*key.indices(len(self)))
        elif isinstance(key, int):
            idx = [key]
        else:
            raise TypeError("Invalid argument type.")

        for i in idx:
            if self.trees[i] is None:
                self.trees[i] = Tree.from_swc(self.swcs[i], **self.read_kwargs)

    @staticmethod
    def from_swc(swc_dir: str, suffix: str = ".swc", **kwargs) -> "Population":
        swcs = Population.find_swcs(swc_dir, suffix)
        population = Population(swcs)
        population.source = swc_dir
        population.read_kwargs = kwargs
        return population

    @staticmethod
    def from_eswc(swc_dir: str, suffix: str = ".eswc", **kwargs) -> "Population":
        kwargs.setdefault("extra_cols", [])
        kwargs["extra_cols"].extend(k for k, t in eswc_cols)
        return Population.from_swc(swc_dir, suffix, **kwargs)

    @staticmethod
    def find_swcs(swc_dir: str, suffix: str = ".swc") -> list[str]:
        """Find all swc files."""
        swcs = list[str]()
        for root, _, files in os.walk(swc_dir):
            files = [f for f in files if os.path.splitext(f)[-1] == suffix]
            swcs.extend([os.path.join(root, f) for f in files])

        return swcs



"""Branch anlysis of tree."""

from functools import cached_property
from typing import List, TypeVar

import numpy as np
import numpy.typing as npt

from ..core import Branch, Tree

__all__ = ["BranchFeatures"]

T = TypeVar("T", bound=Branch)


class BranchFeatures:
    """Analysis bransh of tree."""

    tree: Tree

    def __init__(self, tree: Tree) -> None:
        self.tree = tree

    def get_count(self) -> int:
        return len(self._branches)

    def get_length(self) -> npt.NDArray[np.float32]:
        """Get length of branches."""
        length = [br.length() for br in self._branches]
        return np.array(length, dtype=np.float32)

    def get_tortuosity(self) -> npt.NDArray[np.float32]:
        """Get tortuosity of path."""
        return np.array([br.tortuosity() for br in self._branches], dtype=np.float32)

    def get_angle(self, eps: float = 1e-7) -> npt.NDArray[np.float32]:
        """Get agnle between branches.

        Returns
        -------
        angle : npt.NDArray[np.float32]
            An array of shape (N, N), which N is length of branches.
        """

        return self.calc_angle(self._branches, eps=eps)

    @staticmethod
    def calc_angle(branches: List[T], eps: float = 1e-7) -> npt.NDArray[np.float32]:
        """Calc agnle between branches.

        Returns
        -------
        angle : npt.NDArray[np.float32]
            An array of shape (N, N), which N is length of branches.
        """

        vector = np.array([br[-1].xyz() - br[0].xyz() for br in branches])
        vector_dot = np.matmul(vector, vector.T)
        vector_norm = np.linalg.norm(vector, ord=2, axis=1, keepdims=True)
        vector_norm_dot = np.matmul(vector_norm, vector_norm.T) + eps
        arccos = np.clip(vector_dot / vector_norm_dot, -1, 1)
        angle = np.arccos(arccos)
        return angle

    @cached_property
    def _branches(self) -> List[Tree.Branch]:
        return self.tree.get_branches()



"""Transformation in tree."""

from typing import Any, Generic, TypeVar, cast, overload

__all__ = ["Transform", "Transforms", "Identity"]

T, K = TypeVar("T"), TypeVar("K")

T1, T2, T3 = TypeVar("T1"), TypeVar("T2"), TypeVar("T3")  # pylint: disable=invalid-name
T4, T5, T6 = TypeVar("T4"), TypeVar("T5"), TypeVar("T6")  # pylint: disable=invalid-name


class Transform(Generic[T, K]):
    r"""An abstract class representing a :class:`Transform`.

    All transforms that represent a map from `T` to `K`.

    Methods
    -------
    __call__(x: T) -> K
        All subclasses should overwrite :meth:`__call__`, supporting
        applying transform in `x`.
    __repr__() -> str
        Subclasses could also optionally overwrite :meth:`__repr__`.
        Avoid using the underscore `_` because it is used by
        `Transforms`. If not provied, class name will be a default
        value.
    """

    def __call__(self, x: T) -> K:
        """Apply transform."""
        raise NotImplementedError()

    def __repr__(self) -> str:
        return self.__class__.__name__


class Transforms(Transform[T, K]):
    """A simple typed wrapper for transforms."""

    transforms: list[Transform[Any, Any]]

    # fmt:off
    @overload
    def __init__(self, t1: Transform[T, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, T2],
                       t3: Transform[T2, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, T2],
                       t3: Transform[T2, T3], t4: Transform[T3, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, T2],
                       t3: Transform[T2, T3], t4: Transform[T3, T4],
                       t5: Transform[T4, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, T2],
                       t3: Transform[T2, T3], t4: Transform[T3, T4],
                       t5: Transform[T4, T5], t6: Transform[T5, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, T2],
                       t3: Transform[T2, T3], t4: Transform[T3, T4],
                       t5: Transform[T4, T5], t6: Transform[T5, T6],
                       t7: Transform[T6, Any], /, *transforms: Transform[Any, Any]) -> None: ...
    # fmt:on
    def __init__(self, *transforms: Transform[Any, Any]) -> None:
        self.transforms = list(transforms)

    def __call__(self, x: T) -> K:
        """Apply transforms."""
        for transform in self.transforms:
            x = transform(x)

        return cast(K, x)

    def __getitem__(self, idx: int) -> Transform[Any, Any]:
        return self.transforms[idx]

    def __repr__(self) -> str:
        return "_".join([str(transform) for transform in self])

    def __len__(self) -> int:
        return len(self.transforms)


class Identity(Transform[T, T]):
    """Resurn input as-is."""

    def __call__(self, x: T) -> T:
        return x

    def __repr__(self) -> str:
        return ""



"""Painter utils."""

import os
import weakref
from typing import Any, Dict, Literal, Tuple

import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ..core import SWCLike, Tree
from ..utils import (
    draw_lines,
    draw_xyz_axes,
    get_fig_ax,
    palette,
    Camera,
    Vector3D,
)

__all__ = ["draw"]

CameraPreset = Literal["xy", "yz", "zx", "yx", "zy", "xz"]
CameraPresets: Dict[CameraPreset, Camera] = {
    "xy": Camera((0, 0, 0), (0, 0, -1), (0, 1, 0)),
    "yz": Camera((0, 0, 0), (-1, 0, 0), (0, 0, 1)),
    "zx": Camera((0, 0, 0), (0, -1, 0), (1, 0, 0)),
    "yx": Camera((0, 0, 0), (0, 0, -1), (0, -1, 0)),
    "zy": Camera((0, 0, 0), (-1, 0, 0), (0, 0, -1)),
    "xz": Camera((0, 0, 0), (0, -1, 0), (-1, 0, 0)),
}
CameraOptions = (
    Vector3D | Tuple[Vector3D, Vector3D] | Tuple[Vector3D, Vector3D, Vector3D]
)

ax_weak_dict = weakref.WeakKeyDictionary[Axes, Dict[str, Any]]({})


def draw(
    swc: SWCLike | str,
    *,
    fig: Figure | None = None,
    ax: Axes | None = None,
    camera: CameraOptions | CameraPreset = "xy",
    color: Dict[int, str] | str | None = None,
    label: str | Literal[True] = True,  # TODO: support False
    **kwargs,
) -> tuple[Figure, Axes]:
    """Draw neuron tree.

    Parameters
    ----------
    swc : SWCLike | str
        If it is str, then it is treated as the path of swc file.
    fig : ~matplotlib.axes.Figure, optional
    ax : ~matplotlib.axes.Axes, optional
    camera : CameraOptions | CameraPreset, default "xy"
        Camera options (position, look-at, up). One, two, or three
        vectors are supported, if only one vector, then threat it as
        look-at, so camera is ((0, 0, 0), look-at, (0, 1, 0));if two
        vector, then then threat it as (look-at, up), so camera is
        ((0, 0, 0), look-at, up). An easy way is to use the presets
        "xy", "yz" and "zx".
    color : Dict[int, str] | "vaa3d" | str, optional
        Color map. If is dict, segments will be colored by the type of
        parent node.If is string, the value will be use for any type.
    label : str | bool, default True
        Label of legend, disable if False.
    **kwargs : dict[str, Unknown]
        Forwarded to `~matplotlib.collections.LineCollection`.

    Returns
    -------
    fig : ~matplotlib.axes.Figure
    ax : ~matplotlib.axes.Axes
    """

    fig, ax = get_fig_ax(fig, ax)
    ax_weak_dict.setdefault(ax, {})
    ax_weak_dict[ax].setdefault("swc", [])

    if isinstance(swc, str):
        swc = Tree.from_swc(swc)
    ax_weak_dict[ax]["swc"].append(swc)

    my_camera = get_camera(camera)
    my_color = get_color(ax, swc, color)

    xyz = swc.xyz()
    starts, ends = swc.id()[1:], swc.pid()[1:]
    lines = np.stack([xyz[starts], xyz[ends]], axis=1)

    collection = draw_lines(lines, ax=ax, camera=my_camera, color=my_color, **kwargs)

    # legend
    ax_weak_dict[ax].setdefault("handles", [])
    ax_weak_dict[ax]["handles"].append(collection)
    set_lable(ax, swc, label)

    ax.autoscale()

    if len(ax_weak_dict[ax]["swc"]) == 1:
        ax.set_aspect(1)
        ax.spines[["top", "right"]].set_visible(False)
        ax.text(0.05, 0.95, r"$\mu m$", transform=ax.transAxes)
        draw_xyz_axes(ax=ax, camera=my_camera)
    else:
        # legend
        handles = ax_weak_dict[ax].get("handles", [])
        labels = ax_weak_dict[ax].get("labels", [])
        ax.legend(handles, labels, loc="upper right")

    return fig, ax


def get_camera(camera: CameraOptions | CameraPreset) -> Camera:
    if isinstance(camera, str):
        return CameraPresets[camera]

    if len(camera) == 1:
        return Camera((0, 0, 0), camera, (0, 1, 0))

    if len(camera) == 2:
        return Camera((0, 0, 0), camera[0], camera[1])

    return Camera(*camera)


def get_color(
    ax: Axes, swc: SWCLike, color: Dict[int, str] | str | None
) -> str | list[str]:
    if color == "vaa3d":
        color = palette.vaa3d

    if isinstance(color, str):
        return color

    # choose default
    ax_weak_dict[ax].setdefault("color", -1)
    ax_weak_dict[ax]["color"] += 1
    c = palette.default[ax_weak_dict[ax]["color"] % len(palette.default)]

    if isinstance(color, dict):
        types = swc.type()[:-1]  # colored by type of parent node
        return list(map(lambda type: color.get(type, c), types))

    return c


def set_lable(ax: Axes, swc: SWCLike, label: str | bool):
    ax_weak_dict[ax].setdefault("labels", [])
    if label is False:
        ax_weak_dict[ax]["labels"].append(False)
        return

    if label is True:
        try:
            (_, tail) = os.path.split(swc.source)
            label = tail
        except:  # type: ignore
            label = swc.source

    ax_weak_dict[ax]["labels"].append(label)



"""Debug helpers"""

from functools import wraps
import time

__all__ = ["func_timer"]


def func_timer(function):
    @wraps(function)
    def function_timer(*args, **kwargs):
        print(f"[Function: {function.__name__} start...]")
        t0 = time.time()
        result = function(*args, **kwargs)
        t1 = time.time()
        print(f"[Function: {function.__name__} finished, spent time: {t1 - t0:.2f}s]")
        return result

    return function_timer



"""Transformations in branch requires pytorch."""

from typing import Callable, Literal

import numpy as np
import numpy.typing as npt
import torch

from ...core import Branch
from ...transforms import Transform

__all__ = ["BranchToTensor"]


class BranchToTensor(Transform[Branch, torch.Tensor]):
    """Transform branch to ~torch.Tensor."""

    channels: str
    channel_first: bool

    get_channels: Callable[[Branch], npt.NDArray[np.float32]]

    def __init__(
        self, channels: Literal["xyz", "xyzr"] = "xyz", channel_first: bool = True
    ):
        """Transform branch to ~torch.Tensor.

        Parameters
        ----------
        channels : str, default `xyz`
            Output channels and order, support `xyz` and `xyzr` now.
        channel_first : bool
            Tensor dimension order. If `True`, the tensor is of shape (C, N), otherwise of
            shape (N, C).
        """

        super().__init__()
        self.channels = channels
        self.channel_first = channel_first

        match self.channels:
            case "xyz":
                self.get_channels = lambda x: x.xyz()
            case "xyzr":
                self.get_channels = lambda x: x.xyzr()
            case _:
                raise ValueError("unsupported channels.")

    def __repr__(self) -> str:
        return f"BranchToTensor-{self.channels}{'-ChannelFirst' if self.channel_first else ''}"

    def __call__(self, x: Branch) -> torch.Tensor:
        channels = self.get_channels(x)  # (N, C)
        tensor = torch.from_numpy(channels).float()
        return tensor.T if self.channel_first else tensor



"""Neuron trees."""

from .branch import *
from .branch_tree import *
from .node import *
from .path import *
from .population import *
from .swc import *
from .tree import *
from .tree_utils import *



"""Transformation in tree."""

from typing import Callable

import numpy as np

from ..core import BranchTree, Tree, cut_tree
from .base import Transform

__all__ = ["ToBranchTree", "TreeNormalizer", "CutByBifurcationOrder"]


class ToBranchTree(Transform[Tree, BranchTree]):
    """Transform tree to branch tree."""

    def __call__(self, x: Tree) -> BranchTree:
        return BranchTree.from_tree(x)


class TreeNormalizer(Transform[Tree, Tree]):
    """Noramlize coordinates and radius to 0-1."""

    def __call__(self, x: Tree) -> "Tree":
        """Scale the `x`, `y`, `z`, `r` of nodes to 0-1."""
        new_tree = x.copy()
        for key in ["x", "y", "z", "r"]:  # TODO: does r is the same?
            v_max = np.max(new_tree.ndata[key])
            v_min = np.min(new_tree.ndata[key])
            new_tree.ndata[key] = (new_tree.ndata[key] - v_min) / v_max

        return new_tree


class CutByBifurcationOrder(Transform[Tree, Tree]):
    """Cut tree by bifurcation order"""

    max_bifurcation_order: int
    enter: Callable[[Tree.Node, int | None], tuple[int, bool]]

    def __init__(self, max_bifurcation_order: int) -> None:
        self.max_bifurcation_order = max_bifurcation_order

        def enter(n: Tree.Node, parent_level: int | None) -> tuple[int, bool]:
            if parent_level is None:
                level = 0
            elif n.is_bifurcation():
                level = parent_level + 1
            else:
                level = parent_level
            return (level, level >= self.max_bifurcation_order)

        self.enter = enter

    def __call__(self, x: Tree) -> Tree:
        return cut_tree(x, enter=self.enter)

    def __repr__(self) -> str:
        return f"CutByBifurcationOrder-{self.max_bifurcation_order}"



"""Tree Folder Dataset."""

from typing import Generic, TypeVar, cast

import torch.utils.data

from ...core import Population, Tree
from ...transforms import Identity, Transform

__all__ = ["TreeFolderDataset"]

T = TypeVar("T")
identity = Identity[Tree]()


class TreeFolderDataset(torch.utils.data.Dataset, Generic[T]):
    """Tree folder in swc format."""

    population: Population
    transform: Transform[Tree, T]

    def __init__(
        self,
        swc_dir: str,
        transform: Transform[Tree, T] = identity,
    ) -> None:
        """Create tree dataset.

        Parameters
        ----------
        swc_dir : str
            Path of SWC file directory.
        transfroms : Transfroms[Tree, T], optional
            Branch transformations.

        See Also
        --------
        ~swcgeom.data.torch.transforms : module
            Preset transform set.
        """
        super().__init__()
        self.population = Population.from_swc(swc_dir)
        self.transform = transform

    def __getitem__(self, idx: int) -> T:
        """Get a tree."""
        tree = self.population[idx]
        x = self.transform(tree) if self.transform is not identity else tree
        return cast(T, x)

    def __len__(self) -> int:
        """Get length of set of trees."""
        return len(self.population)
